<?php
/**
 * The admin-specific functionality of the plugin.
 *
 * @link       http://www.contasimple.com
 * @since      1.0.0
 *
 * @package    contasimple
 * @subpackage contasimple/admin
 * @author     Contasimple S.L. <soporte@contasimple.com>
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}

/**
 * Class for the Admin backoffice
 *
 * Defines the plugin name, version, two examples hooks for how to
 * enqueue the admin-specific stylesheet and JavaScript, and other admin logic.
 *
 * @since      1.0.0
 */
class Contasimple_Admin {

	/**
	 * Contasimple Service
	 *
	 * @var        CSService $cs
	 *
	 * @since      1.0.0
	 */
	private $cs;

	/**
	 * Config class
	 *
	 * @var        CSConfig $config
	 *
	 * @since      1.0.0
	 */
	private $config;

	/**
	 * File logger to catch errors
	 *
	 * @var        CSLogger $logger
	 *
	 * @since      1.0.0
	 */
	protected $logger;

	/**
	 * Getter for the config class
	 *
	 * @return     mixed
	 *
	 * @since      1.0.0
	 */
	public function get_config() {

		return $this->config;
	}

	/**
	 * Setter for the config class
	 *
	 * @param      mixed $config Instance of the Config class to store.
	 *
	 * @since      1.0.0
	 */
	public function set_config( $config ) {

		$this->config = $config;
	}

	/**
	 * Load the config values from WP Options table
	 *
	 * @since      1.0.0
	 */
	protected function load_config_values() {

		$this->config = get_option( 'contasimple_settings_account' );

		if ( empty( $this->config ) ) {
			$this->config = new CSConfig();
		}
	}

	/**
	 * Store changed values back into ps_configuration table
	 *
	 * @param    CSConfig $config      The custom Config class object with latest-state user/API details.
	 * @param    string $expire_time API access token expiration time, pass response time + current date to check when
	 *                               to request token refresh.
	 *
	 *@since    1.0.0
	 */
	public function store_config_values( $config, $expire_time = null ) {
		$this->config = $config;

		if ( ! empty( $expire_time ) ) {
			$this->config->setExpireTime( $expire_time );
		}

		// Serializes all data into a single wp_options field.
		update_option( 'contasimple_settings_account', $this->config );
	}

	/**
	 * Getter for Contasimple settings as WP Options
	 *
	 * @return     mixed|void
	 *
	 * @since      1.0.0
	 */
	public static function get_config_static() {

		return get_option( 'contasimple_settings_account' );
	}

	/**
	 * Getter for the Service class
	 *
	 * @return     CSService
	 *
	 * @since      1.0.0
	 */
	public function get_service() {

		return $this->cs;
	}

	/**
	 * Init the service class
	 *
	 * @since      1.0.0
	 */
	public function init_service() {

		$this->cs = new CSService( $this->get_config() );
		$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );
	}

	/**
	 * Getter for the Logger class
	 *
	 * @return  CSLogger
	 *
	 * @since   1.5
	 */
	public function getLogger() {
		return $this->logger;
	}

	/**
	 * Initialize the class and set its properties.
	 *
	 * @since    1.0.0
	 */
	public function __construct() {

		$this->logger = CSLogger::getDailyLogger();
		$this->load_config_values();
	}

	/**
	 * Register the stylesheets for the admin area.
	 *
	 * @since    1.0.0
	 */
	public function enqueue_styles() {

		wp_enqueue_style( 'bootstrap', plugin_dir_url( __FILE__ ) . 'css/bootstrap-cs.css', array(), '3.3.7' );
		wp_enqueue_style( 'jquery-ui-css' );

		// Seems like after 2.2 it is always loaded.
		if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
			wp_enqueue_style( 'woocommerce_admin_styles', WC()->plugin_url() . '/assets/css/admin.css', array(), WC_VERSION );
		}

		$deps = array( 'woocommerce_admin_styles' );
		wp_enqueue_style( Contasimple::$plugin_name, plugin_dir_url( __FILE__ ) . 'css/contasimple-admin.css', $deps, Contasimple::$version, 'all' );
	}

	/**
	 * Register the JavaScript for the admin area.
	 *
	 * @since    1.0.0
	 */
	public function enqueue_scripts() {

		if ( $this->check_bootstrap_loaded() == false ) {
			wp_enqueue_script( 'bootstrap', plugin_dir_url( __FILE__ ) . 'js/bootstrap.min.js', array( 'jquery' ), '3.3.7', true );
		}

		wp_enqueue_script( 'cs_gtag', '//www.googletagmanager.com/gtag/js?id=UA-9928674-21#asyncload', array() );
		wp_enqueue_script( 'cs_ga_analytics', plugin_dir_url( __FILE__ ) . 'js/analytics.js' );
		wp_enqueue_script( Contasimple::$plugin_name, plugin_dir_url( __FILE__ ) . 'js/contasimple-configuration.js', array(
			'jquery',
			'jquery-ui-datepicker',
			'jquery-tiptip',
		), Contasimple::$version, false );
		wp_enqueue_script( Contasimple::$plugin_name . '_orders', plugin_dir_url( __FILE__ ) . 'js/contasimple-orders.js', array(
			'jquery',
			'jquery-ui-tooltip',
		), Contasimple::$version, false );

		// Localize the script with new data.
		$translation_array = array(
			'msg_log_not_found' => __( 'There are no registered logs for the selected day.', 'contasimple' ),
			'msg_no_active_companies_found' => __('You do not have access to any company. You must create a company in Contasimple before linking the plugin for WooCommerce.', 'contasimple')
		);

		wp_localize_script( Contasimple::$plugin_name, 'js_translations', $translation_array );
		wp_localize_script( Contasimple::$plugin_name . '_orders', 'js_translations', $translation_array );
	}

	/**
	 * Check if bootstrap js is already loaded
	 * Note: This method is not infallible.
	 *
	 * @since 1.5
	 *
	 * @return bool True if bootstrap is already loaded, false otherwise.
	 */
	public function check_bootstrap_loaded() {

		global $wp_scripts;

		$bootstrap_enqueued = false;

		foreach ( $wp_scripts->registered as $script ) {
			if ( ( stristr( $script->src, 'bootstrap.min.js') != false ||
			       stristr( $script->src, 'bootstrap.js') != false ||
			       stristr( $script->src, 'bootstrap-modal.js') != false ||
			       stristr( $script->src, 'bootstrap-modal.min.js') != false
			    ) &&
			    wp_script_is( $script->handle, $list = 'enqueued' )
			) {
				$bootstrap_enqueued = true;

				if ( !empty( $this->logger ) ) {
					$this->logger->log('Skipped loading CS own bootstrap.js file. Detected potential conflict with bootstrap js: [' . $script->handle . '] => ' . $script->src . '.' );
				}

				break;
			}
		}

		return $bootstrap_enqueued;
	}

	public function add_async_to_gtag( $url ) {

		if ( strpos( $url, '#asyncload' ) === false )
			return $url;
		else
			return str_replace( '#asyncload', '', $url )."' async='async";
	}

	/**
	 * Login to CS API via APIKEY
	 *
	 * This method is called via AJAX and returns an AJAX response.
	 * If login is successful, will allow the user to continue with the wizard. Otherwise it will be locked in the
	 * configuration screen.
	 *
	 * @since    1.0.0
	 */
	public function cs_login() {

		global $woocommerce;

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		$result = array(
			'message' => __( 'Connection successful', 'contasimple' ),
		);

		try {
			$this->check_ajax_wizard_allowed();

			$apikey = esc_attr( $_REQUEST['apikey'] ); // Input var okay.
			$user_agent = 'ContasimpleWooCommerce/' . Contasimple::$version . '-' . $woocommerce->version;

			$this->force_login();
			$this->cs = new CSService( $this->config );
			$response = $this->cs->connect( $apikey, $user_agent );

			$this->logger->log( 'Connection with APIKEY successful' );
			$this->store_config_values( $this->cs->getConfig(), $response->getExpiresIn() + time() );

			$companies_list = $this->cs->getCompanies();

			foreach ( $companies_list as $company ) {
				// $currency              = get_woocommerce_currencies()[ $company->getCountry()->getCurrency()->getShortName() ];
				$result['companies'][] = array(
					'id_option'    => $company->getId(),
					'name'         => $company->getDisplayName(),
					// 'currency'     => $currency,
					'country'      => $company->getCountry()->getName(),
					'fiscalRegion' => $company->getFiscalRegion()->getName(),
				);
			}
		} catch ( Exception $e ) {
			if ( $e->getMessage() === 'key is required.' ) {
				$result = array(
					'error'   => true,
					'message' => __( 'Please, enter a valid API Key', 'contasimple' ),
				);
			} elseif ( $e->getMessage() == 'Client is not registered in the system.' ) {
				$result = array(
					'error'   => true,
					'message' => __( 'Your API Key cannot be found in our system. Please, make sure it is correct. If you are still not registered, please register first in Contasimple.com', 'contasimple' ),
				);
			} elseif ( $e->getMessage() === 'invalid_grant' ) {
				$this->force_login();
				$result = array(
					'error'   => true,
					'message' => __( 'Invalid grant', 'contasimple' ),
				);
			} elseif ( strpos( $e->getMessage(), 'Could not resolve' ) !== false ) {
				$result = array(
					'error'   => true,
					'message' => __( 'Could not resolve host. Please check that your server has internet access.', 'contasimple' ),
				);
			} else {
				// Captured an unknown error, log.
				$this->logger->logTransactionEnd( 'Unexpected exception thrown: ' . $e->getMessage() );

				$this->force_login();
				$result = array(
					'error'   => true,
					'message' => __( 'An unknown error occurred. Please try again, and if it persists, please contact Contasimple.', 'contasimple' ),
				);
			}
		}

		$this->logger->logTransactionEnd();
		echo wp_json_encode( $result );

		wp_die(); // this is required to terminate immediately and return a proper response.
	}

	/**
	 * Add NIF to edit order screen
	 *
	 * @param WC_Order $order The order to get the customer NIF for.
	 *
	 * @since    1.0.0
	 */
	public function add_custom_field_nif_checkout_edit_order( $order ) {

		$config_company_identifier_name = $this->get_company_identifier_name();

		if ( !empty( $config_company_identifier_name ) ) {
			$nif_label = sprintf( __('%s / Company identifier', 'contasimple'), $config_company_identifier_name );
		} else {
			$nif_label = __( 'Company identifier', 'contasimple' );
		}

		echo '<p><strong>' . $nif_label . ':</strong> ' . esc_attr( get_post_meta( $order->get_order_number(), 'NIF', true ) ) . '</p>';
	}

	/**
	 * Resume sync queue when an order is completed
	 *
	 * Complete here means that payment also has been collected. This is typically considered an 'invoiceable' state.
	 *
	 * @param int $order_id The order id.
	 *
	 * @since    1.0.0
	 */
	public function order_status_completed( $order_id ) {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		if ( ! empty( $order_id ) ) {

			$order_invoices = CS_Invoice_Sync::get_invoices_from_order( $order_id );

		    if ( ! $this->invoice_can_be_added( $order_invoices, false ) ) {
                $this->resume_queue( false );
            } else {
                $contasimple_invoice_sync = CS_Invoice_Sync::create_empty( $order_id );

                if ( ! empty( $contasimple_invoice_sync ) ) {
                    if ( self::properly_configured() ) {
                        $contasimple_invoice_sync->companyID = $this->config->getCompanyId();
                        $contasimple_invoice_sync->mask      = $this->get_mask_from_order( $order_id );
                        $contasimple_invoice_sync->save();
                        $this->resume_queue( false );
                    }
                } else {
                    $this->logger->log( 'Error creating an entry into the database for the new order.' );
                }
            }
		} else {
			$this->logger->log( 'Empty order_id found on status_completed hook.' );
		}
	}

	/**
	 * Create an invoice for a refund
	 *
	 * AKA order slip or credit note, 'factura simplificada' in Spain
	 *
	 * @param    int   $refund_get_id Refund id passed down by the hook.
	 * @param    array $args          Data about the refund post.
	 *
	 * @since    1.0.0
	 */
	public function refund_created( $refund_get_id, $args ) {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		if ( ! empty( $args['order_id'] ) ) {

            $order_invoices = CS_Invoice_Sync::get_invoices_from_order($args['order_id']);

			// Improve this method, does not contemplate many scenarios.
            if ( ! $this->invoice_can_be_added( $order_invoices, true ) ) {
                $this->resume_queue( false );
            } else {
                $contasimple_invoice_sync = CS_Invoice_Sync::create_empty( $args['order_id'], $refund_get_id, $args );

                if ( ! empty( $contasimple_invoice_sync ) ) {
                    if ( self::properly_configured() ) { // TODO !this->accessAllowed() verify ajax legitimity.
                        $contasimple_invoice_sync->companyID = $this->config->getCompanyId();
                        $contasimple_invoice_sync->mask      = $this->get_mask_from_order( $args['order_id'] );
                        $contasimple_invoice_sync->save();
                        $this->resume_queue( false );
                    }
                } else {
                    $this->logger->log( 'Error creating an entry into the database for the refund.' );
                }
            }
		} else {
			$this->logger->log( 'Empty order_id found on refund_created hook.' );
		}
	}

	/**
	 * Get the number formatting mask to apply for document sequencing
	 *
	 * We have 3 possible choices for the mask.
	 * The mask selected will depend on the customer order info provided.
	 *
	 * If it is a refund, we will always return a mask for the 'CREDIT NOTES' series (Also known as order slips, etc).
	 * If it is a completed order or equivalent, we will return either:
	 * - A mask for regular 'INVOICES' series (those that have valid fiscal information, AKA ADDRESS AND COMPANY ID/NIF/VAT)
	 * - A mask for a 'RECEIPT' series. (Also known as ticket or 'factura simplificada', since Spain requires a whole
	 *   different series if an invoice cannot be generated due to MISSING EITHER ADDRESS OR ID/NIF/VAT).
	 *
	 * @param int $order_id Order class ID to extract the needed data (NIF, Address is typically enforced by default in WC).
	 *
	 * @return string The mask for the specific type of document that we need to create.
	 *
	 * @since    1.0.0
	 */
	public function get_mask_from_order( $order_id ) {

		if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
			try {
                $order = new WC_Order($order_id);
            }
            catch(\Exception $ex){
                $order = new WC_Order();
            }
			$order_status = $order->status;
		} else {
			$order        = wc_get_order( $order_id );
			$order_status = $order->get_status();
		}

		if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
			$is_partial_refund = false;
		} else {
			$is_partial_refund = count( $order->get_refunds() ) > 0 ? true : false;
		}

		$nif     = get_post_meta( $order_id, 'NIF', true );
		$options = get_option( 'woocommerce_integration-contasimple_settings' );

		if ( version_compare( WC_VERSION, '2.7', '<' ) ) {
			$customer_country = $order->billing_country;
		} else {
			$customer_country = $order->get_billing_country();
		}

		if ( 'refunded' === $order_status || $is_partial_refund ) {
			if ( ! empty( $options['refunds_mask'] ) ) {
				$mask = $options['refunds_mask'];
			} else {
				$mask = 'WC-NOTE-AAAA-######';
			}
		} elseif ( ! empty( $nif ) && $this->is_valid_nif( $nif, $customer_country ) ) {
			if ( ! empty( $options['invoices_mask'] ) ) {
				$mask = $options['invoices_mask'];
			} else {
				$mask = 'WC-INV-AAAA-######';
			}
		} else {
			if ( ! empty( $options['receipts_mask'] ) ) {
				$mask = $options['receipts_mask'];
			} else {
				$mask = 'WC-TICK-AAAA-######';
			}
		}

		return $mask;
	}

	/**
	 * Perform actions on order change (if needed be)
	 *
	 * @param   int    $order_id
	 * @param   string $old_status
	 * @param   string $new_status
	 *
	 * @deprecated
	 */
	public function order_status_changed( $order_id, $old_status, $new_status ) {

		if ( 'completed' === $new_status ) {
			// TODO Do we need to really do something if an order changes or not?
		}

		if ( 'refunded' === $new_status ) {
			// WC prior to 2.2 does not have the refund_created hook, so we must act here.
			if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
				$args = array();
				$args['order_id'] = $order_id;
				$this->refund_created(null, $args);
			}
		}
	}

	/**
	 * Sync invoice via AJAX
	 *
	 * This handles syncing at the individual level as an action button on the invoices admin table list.
	 *
	 * @deprecated  We removed this functionality in favour of an automatic queue process. User will not be able to sync
	 *              an individual invoice on request.
	 */
	public function cs_sync() {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		$order_id = (int) $_POST['order_id'];
		$nonce    = esc_attr( $_POST['nonce'] );
		$action   = esc_attr( $_POST['action'] );

		if ( ! empty( $order_id ) ) {
			$contasimple_invoice_sync = CS_Invoice_Sync::get_from_order( $order_id );
			if ( $this->access_allowed( $nonce, $action ) ) {
				// TODO Ported from PS but this is not being used in WC, review and think about keeping or removing.
				$contasimple_invoice_sync->attempts = 0; // Reset because is manually.
				// Before sending, make sure we are sending to the current CS company.
				$this->maybe_update_cs_company_settings();
				$response = $this->create_or_update_invoice( $contasimple_invoice_sync );
			} else {
				$this->logger->logTransactionEnd( 'Access not allowed. Timed out, invalid user requesting or configuration info is missing.' );
				$response = array(
					'redirect' => admin_url( 'edit.php?post_type=shop_order' ),
				);
			}
		} else {
			$this->logger->log( 'Empty order_id found when receiving AJAX request on the client panel.' );
			$response = array(
				'error'   => true,
				'message' => __( 'Error trying to retrieve order data from AJAX request.', 'contasimple' ),
			);
		}

		$response['summary'] = $this->get_summary( null, true );

		echo wp_json_encode( $response );

		exit();
	}

	/**
	 * Send individual order info to Contasimple
	 *
	 * @param   ContasimpleInvoice $contasimple_invoice_sync An entity that keeps syncing info about this order.
	 *
	 * @return  array A response with info about the result
	 *
	 * @since    1.0.0
	 */
	public function create_or_update_invoice( $contasimple_invoice_sync ) {

		$order = null;
		$order_data = null;
		$cs_invoice = null;
		$cs_invoice_post = null;
		$type = 'invoice';

		try {
			$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

			// Retrieve WC order object from the ID stored in our sync table row.
			$order = Contasimple_WC_Backward_Compatibility::get_order_from_id( $contasimple_invoice_sync->order_id );

			// Order could have been deleted or fail for whatever reason unknown to us, abort right now.
			if ( empty( $order ) ) {
				// The exception will be caught in the main sync loop and mark the invoice with the pertinent error
				// code.
				throw new \Exception( '', MISSING_ORDER );
			}

			// Only for logging purposes.
			$order_data = Contasimple_WC_Backward_Compatibility::get_order_data( $order );

			$order_status = Contasimple_WC_Backward_Compatibility::get_order_status( $order );

			// The data regarding the sync process, we will update with either SYNC or an ERROR code + message.
			$cs_invoice_post = get_post( $contasimple_invoice_sync->ID );

			// Custom field stored from the checkout page, allows us to sync customer info.
			$nif = get_post_meta( $contasimple_invoice_sync->order_id, 'NIF', true );

			// Get the currency symbol of the WC order paid amount, ie: $ or €
			$currency = Contasimple_WC_Backward_Compatibility::get_currency( $order );
			$currency_symbol = html_entity_decode( get_woocommerce_currency_symbol( $currency ) );

			// TODO Review this, as the order completed date might not be really suitable for a invoice creation date.
			// Completed date happens when the order is marked as completed.
			$date = Contasimple_WC_Backward_Compatibility::get_order_completed_date( $order );

			// Should never happen.
			if ( empty( $date ) ) {
				throw new \Exception( '', INVALID_DATE_COMPLETED );
			}

			// Format the CS period based on the completed date, ie: 2019-4T
			$period = Contasimple_CPT::get_period_from_date( $date );

			// We will use this to generate CS negative amounts by just multiplying the order items qty by 1 or -1.
			$sign = Contasimple_WC_Backward_Compatibility::get_invoice_sign_from_order( $order, $contasimple_invoice_sync->args);

			$type = $sign < 0 ?
				'refund' : 'invoice';

			// Since we keep the companyID inside the synced invoice data in the meta table, we have to make sure that
			// if the company has changed, the current sync reflects this (the rule used here is that the company who
			// syncs the invoice gets the authorship).
			$contasimple_invoice_sync->companyID = $this->get_config()->getCompanyId();

			// Store currencies, we will check that both match before syncing.
			$contasimple_invoice_sync->cs_currency = $this->get_config()->getCurrencySymbol();
			$contasimple_invoice_sync->order_currency = $currency_symbol;

			add_post_meta( $contasimple_invoice_sync->ID, 'cs_currency', $contasimple_invoice_sync->cs_currency);
			add_post_meta( $contasimple_invoice_sync->ID, 'order_currency', $contasimple_invoice_sync->order_currency);

			// Unless the state indicates an error with a payment or an invoice state change, we will try to sync.
			if ( PAYMENT_SYNC_ERROR  !== (int) $contasimple_invoice_sync->state &&
			                 SYNC_OK !== (int) $contasimple_invoice_sync->state &&
				             CHANGED !== (int) $contasimple_invoice_sync->state
			) {
				if ( empty( $contasimple_invoice_sync->number ) ) {
					$this->logger->log( 'Retrieving next invoice number...' );
					$this->cs = new CSService( $this->get_config() );
					$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );
					$contasimple_invoice_sync->number = $this->cs->getNextInvoiceNumber( $period, $contasimple_invoice_sync->mask );
				}

				$contasimple_invoice_sync->update_sync_state( null, PENDING, null, $contasimple_invoice_sync->attempts + 1 );
				$this->logger->log( 'Internal sync status of object type ' . $type . ' updated to PENDING' );
				$this->logger->log( sprintf( 'Gathering data for invoice number [%1s] from period [%2s]... ', $contasimple_invoice_sync->number, $period ) );

				$attempt = 1;

				while ( $attempt <= 3 ) {
					// Attempt 1 tries syncing WC invoice line per line exactly.
					// Attempt 2 will try to group invoice lines per VAT type to avoid rounding issues.
					// Attempt 3 will desperately try to sync WC order final total amount and VAT amount only
					// if everything else failed.
					$this->logger->log( 'CS Invoice calculation strategy #' . $attempt );
					$cs_invoice = $this->convert_order_to_cs_invoice( $order, $contasimple_invoice_sync, $nif, $date, $sign, $attempt );

					$args = $contasimple_invoice_sync->args;
					if ( ! empty( $args ) && is_serialized( $args ) ) {
						$args = unserialize( $args );
					}

					if ( $order_status == 'refunded' || ( $order_status == 'completed' && empty( $args ) ) ) {
						$this->logger->log( 'Validating order data against CS invoice... ' );
						$success = $this->validate_invoice_amounts( $order, $cs_invoice, $sign );
					} else {
						// For now skip validation of partial refunds, method needs to be revisited for this scenario.
						$success = 1;
					}

					// If our result either fully matches (1)
					// or it meets the threshold allowed (0)
					// we can proceed with the sync.
					if ( $success >= 0 ) {
						$this->logger->log( 'Sending invoice...' );
						$cs_response = $this->send_invoice( $period, $cs_invoice, 0 );

						$contasimple_invoice_sync->amount = $cs_response->getTotalAmount();
						$this->logger->log( 'Invoice sent OK!' );

						if ( isset( $cs_invoice['warning'] ) ) {
							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), $cs_invoice['warning'], null );
							$this->logger->log( 'Status changed to SYNC_OK (WITH WARNINGS)' );
						} else if ( 0 == $success ) {
							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), SYNCED_WITH_ROUNDING_ISSUES, null );
							$this->logger->log( 'Status changed to SYNC_OK (WITH ROUNDING WARNINGS)' );
						} else if ( 1 == $success && $attempt == 3 ) {
							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), SYNCED_WITH_ROUNDING_ISSUES, null );
							$this->logger->log( 'Status changed to SYNC_OK (WITH ROUNDING WARNINGS)' );
						} else {
							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), SYNC_OK, null );
							$this->logger->log( 'Status changed to SYNC_OK' );
						}

						// If we made it here without throwing an exception, the invoice is synced so next step is sending payments,
						// Unless:
						// - If it is an error scenario (as invoice sync is needed first)
						// - An update (TODO ignore, delete and resend payments, or send missing?).
						if ( 0 < $contasimple_invoice_sync->externalID  &&
						     INVOICE_REPEATED !== (int) $contasimple_invoice_sync->state &&
						     CHANGED !== (int) $contasimple_invoice_sync->state
						) {
							if ( ! empty( $args ) ) {
								$amount = $args['amount'] * $sign;
							} else {
								$amount = $order->get_total() * $sign;
							}

							$method = Contasimple_WC_Backward_Compatibility::get_order_payment_method( $order );

							$this->send_payment( $contasimple_invoice_sync, $period, $method, $amount );

							// Keep the current state so that if it was OK, still is OK, if there was an empty VAT for national warning,
							// the warning is still persisted (payment does not affect this), etc.
							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), $contasimple_invoice_sync->state, null );
						}

						// If there has been changes in the original order, update them in CS as well.
						if ( 0 < (int) $contasimple_invoice_sync->externalID &&
						     CHANGED === (int) $contasimple_invoice_sync->state
						) {
							$this->logger->log( 'Already exists. Updating invoice...' );
							$this->send_invoice( $period, $cs_invoice, $contasimple_invoice_sync->externalID );
							$this->logger->log( 'Invoice updated OK!' );

							$contasimple_invoice_sync->update_sync_state( $cs_response->getId(), SYNC_OK, null );
							$this->logger->log( 'Status changed to SYNC_OK' );
						}

						// At this point we consider 'SYNCED' being the same success state with or without payments
						// Otherwise if a payment exists but is missing then it is a special case of an error state.
						$response = array(
							'message'   => Contasimple_CPT::get_message_html( $contasimple_invoice_sync->state, $contasimple_invoice_sync->api_message ),
							'icon'      => Contasimple_CPT::get_state_html( $contasimple_invoice_sync->state ),
							'buttons'   => Contasimple_CPT::render_action_buttons( $cs_invoice_post, array( 'cs_view', 'cs_pdf' ) ),
							'date_sync' => date( 'Y-m-d H:i:s' ),
						);

						$this->logger->log( 'Everything seems OK! Still, internal data logging is provided, just in case:' );
						$this->logger->log( '  >  wc_invoice = ' . wp_json_encode( $contasimple_invoice_sync, JSON_UNESCAPED_UNICODE ) );
						$this->logger->log( '  >  wc_order   = ' . wp_json_encode( $order_data, JSON_UNESCAPED_UNICODE ) );
						$this->logger->log( '  >  CSInvoice  = ' . wp_json_encode( $cs_invoice, JSON_UNESCAPED_UNICODE ) );

						$this->logger->logTransactionEnd();

						break;
					} else {
						// Try the next strategy
						$attempt ++;
						$this->logger->log( 'Trying next CS Invoice calculation strategy...' );
					}
				}

				if ( $success < 0 ) {
					// We could not find a satisfactory sync strategy. Let the user know and suggest.
					$this->logger->log( 'Could not find appropriate amounts to sync.' );
					throw new \Exception( 'Could not find appropriate amounts to sync.', TOTAL_AMOUNT_INVALID );
				}
			}
		} catch ( \Exception $e ) {
			$this->logger->log( 'An error occurred trying to sync the ' . $type . ' at attempt #' . $contasimple_invoice_sync->attempts . '. Internal data is provided:' );
			$this->logger->log( '  >  wc_invoice = ' . wp_json_encode( $contasimple_invoice_sync, JSON_UNESCAPED_UNICODE ) );
			$this->logger->log( '  >  wc_order   = ' . wp_json_encode( $order_data, JSON_UNESCAPED_UNICODE ) );

			if ( ! empty( $cs_invoice ) ) {
				$this->logger->log( '  >  CSInvoice  = ' . wp_json_encode( $cs_invoice, JSON_UNESCAPED_UNICODE ) );
			}
			$this->logger->log( 'Exception thrown message: ' . $e->getMessage() );
			$this->logger->logTransactionEnd( 'Internal status updated to ERROR (code ' . $e->getCode() . ' => ' . trim( strip_tags( Contasimple_CPT::get_message_html( $e->getCode(), $contasimple_invoice_sync->api_message ) ) ) . ')' );

			$contasimple_invoice_sync->number = '';

			// Only if the error is API-related, save the API message and show this to the end user instead of a generic
			// error message. Note that this is not translated at the Wordpress level and might be shown in a different
			// language depending on how things are set up.
			if ( (int) $e->getCode() === API_ERROR ) {
				$contasimple_invoice_sync->update_sync_state( null, $e->getCode(), $e->getMessage() );
			} else {
				$contasimple_invoice_sync->update_sync_state( null, $e->getCode(), null );
			}

			$response = array(
				'message'   => Contasimple_CPT::get_message_html( $e->getCode(), $contasimple_invoice_sync->api_message ),
				'icon'      => Contasimple_CPT::get_state_html( $e->getCode() ),
				'buttons'   => Contasimple_CPT::render_action_buttons( $cs_invoice_post, array( 'cs_stop' ) ),
				'date_sync' => date( 'Y-m-d H:i:s' ),
				'error'     => true,
			);
		}

		if ( ! isset( $response['error'] ) ) {
			$this->maybe_send_email_and_update_status( $order, $contasimple_invoice_sync, false );
		}

		return $response;
}

	/**
	 * Send invoice email to customer if configured to automatically send (see WC Emails docs)
	 *
	 * @param    WC_Order           $order                    The order to email about.
	 * @param    ContasimpleInvoice $contasimple_invoice_sync The entity that holds the invoice sync info.
	 * @param    bool               $manual                   True if manual triggered sending process, false if auto.
	 *
	 * @since    1.0.0
	 */
	private function maybe_send_email_and_update_status( $order, $contasimple_invoice_sync, $manual = false ) {

		// We need to force the mailer init so that it can register the hook 'woocommerce_' for our custom invoice email, it does not load by default.
		WC()->mailer();

		$result = apply_filters( 'woocommerce_cs_invoice_generated_sync', $order, $contasimple_invoice_sync, $manual );

		if ( true === $result ) {
			$contasimple_invoice_sync->mail_status = EMAIL_SENT;
			$contasimple_invoice_sync->save();
		} elseif ( false === $result ) {
			$contasimple_invoice_sync->mail_status = EMAIL_FAILED;
			$contasimple_invoice_sync->save();
		} else {
			$contasimple_invoice_sync->mail_status = EMAIL_NOT_SENT;
			$contasimple_invoice_sync->save();
		}
	}

	/**
	 * Resume syncing queue
	 *
	 * TODO: See if we can use WP Background Processing to improve performance and scalability.
	 *
	 * @param $is_manual bool  By default true, used to check if we need to sync again after a certain amount of attempts.
	 *                         Set to false if you want to skip syncing failed invoices.
	 *
	 * @since    1.0.0
	 */
	public function resume_queue( $is_manual = true ) {

		try {
			$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );
			$this->logger->log( 'Hook triggered > After order status changed. Checking if need to auto sync pending invoices...' );

			// If it has been 24h since the last attempt, trigger bulk sync > For now we are syncing always
			if ( true ) { //(!$this->get_config()->getLastAutoSync() || strtotime($this->get_config()->getLastAutoSync()) + (60 * 60 * HOURS_TO_SYNC_AGAIN) < time()) {
				// $this->logger->log( 'More than 24h passed sync last sync. Trying to sync invoices...' );
				$results = $this->send_pending_invoices( $is_manual );

				// Update last sync date on database so that we don't try again until the next HOURS_TO_SYNC_AGAIN defined const (ex: 24)
				$now = new DateTime();
				$this->get_config()->setLastAutoSync( $now->format( 'Y-m-d H:i:s' ) );
				$this->store_config_values( $this->get_config() );

				if ( $results['error'] > 0 ) {
					//$this->logger->logTransactionEnd( 'Errors were detected and the auto-sync process stopped. Auto-sync timer has been reset to look after 24h from now on.' );
					$this->logger->logTransactionEnd( 'Sync process ended with errors.' );
				} else {
					$this->logger->logTransactionEnd( 'Invoices synced OK. Auto-sync timer has been reset to look after 24h from now on.' );
				}
			} else {
				$this->logger->logTransactionEnd( 'Not 24h passed since last automatic sync. Sync skipped.' );
			}
		} catch ( \Exception $e ) {
			$this->logger->logTransactionEnd( 'Exception thrown while trying to autosync pending invoices: ' . $e->getMessage() );
		}
	}

	/**
	 * Resume sync queue
	 *
	 * This method does the same as the resume_queue() but is meant to be called when the user explicitly wants to
	 * trigger the action, not in response of any WP hook like order_completed or refund_created.
	 *
	 * @since 1.4.2
	 */
	public function resume_queue_manually() {
		$this->resume_queue( true );

		// Redirect to cs_invoice post page in order to get rid of query params like sync=resume that trigger
		// accidentally the sync process when not intended, like a user pressing enter or F5 on the page.
		wp_redirect( esc_url( add_query_arg( 'post_type', 'cs_invoice', admin_url( 'edit.php' ) ) ) );
		exit();
	}

	/**
	 * Retrieve pending invoices and try to sync them one by one.
	 *
	 * @param $is_manual
	 *
	 * @return array
	 *
	 * @since    1.0.0
	 */
	public function send_pending_invoices( $is_manual ) {

		set_time_limit( 60 * 5 );

		$results = array(
			'total'        => 0,
			'synced'       => 0,
			'error'        => 0,
			'max_attempts' => 0,
		);

		$contasimple_invoices = CS_Invoice_Sync::get_pending_invoices();

		$results['total'] = count( $contasimple_invoices );

		$this->logger->log( sprintf( 'There are %1d pending invoices.', $results['total'] ) );

		// Before sending, make sure we are sending to the current CS company.
		$this->maybe_update_cs_company_settings();

		// Iterate orders and send one by one.
		foreach ( $contasimple_invoices as $contasimple_invoice ) {
			try {
				if ( $contasimple_invoice->attempts < MAX_SYNC_ATTEMPTS || true === $is_manual ) {
					if ( empty( $contasimple_invoice->mask ) ) {
						// Mask should be set on creation but if it failed there, retry. TODO Right?
						$contasimple_invoice->mask = $this->get_mask_from_order( $contasimple_invoice->order_id );
					}

					$result = $this->create_or_update_invoice( $contasimple_invoice );

					if ( isset( $result['error'] ) ) {
						$results['error'] += 1;
						// Fist attempt and failed, send the email error. Skip for the following ones.
						if ($contasimple_invoice->attempts == 1) {
							$this->send_error_sync_mail( $contasimple_invoice, $result );
						}
					} else {
						$results['synced'] += 1;
					}
				} else {
					$results['max_attempts'] += 1;
				}
			} catch ( \Exception $e ) {
				$results['error'] += 1;
				$this->logger->log( 'Internal status updated to ERROR (code ' . $e->getCode() . ')' );
			}
		}


		return $results;
	}

	/**
	 * Stop syncing a certain order
	 *
	 * @throws Exception If something went wrong.
	 *
	 * @since    1.0.0
	 */
	public function cs_stop() {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		$cs_invoice_id = (int) $_REQUEST['cs_invoice_id']; // Input var okay.
		$nonce         = esc_attr( $_REQUEST['nonce'] ); // Input var okay.
		$action        = esc_attr( $_REQUEST['action'] ); //Input var okay.

		if ( ! $this->access_allowed( $nonce, $action ) ) {
			$this->logger->logTransactionEnd( 'Access not allowed. Configuration info is missing.' );
			$response = array(
				'redirect' => admin_url( 'edit.php?post_type=shop_order' ),
			);
		}

		if ( null !== $cs_invoice_id ) {
			try {
				$this->logger->log( 'Stopping invoice sync with ID [' . $cs_invoice_id . ']' );
				$contasimple_invoice_sync = new CS_Invoice_Sync( $cs_invoice_id );

				if ( $contasimple_invoice_sync->state == PENDING ||
				     $contasimple_invoice_sync->state == NOT_SYNC ||
				     $contasimple_invoice_sync->state >= SYNC_ERROR && '0' === $contasimple_invoice_sync->externalID ) {
					wp_delete_post( $cs_invoice_id, true );
					$this->logger->logTransactionEnd( 'Successfully deleted from sync queue.' );
					$contasimple_cpt = new Contasimple_CPT();
					$response = array(
						//'redirect' => admin_url( 'edit.php?post_type=cs_invoice' ), // recover old ajax behaviour
						'summary' => $contasimple_cpt->get_summary(null, true)
					);
				} else {
					throw new Exception();
				}
			} catch ( \Exception $e ) {
				$this->logger->log( 'Error updating the internal state of the item. Internal data is provided:' );
				$this->logger->log( '  >  Exception message = ' . $e->getMessage() );
				$this->logger->logTransactionEnd();

				$cs_invoice_post = get_post( $cs_invoice_id );

				$response = array(
					'error'   => true,
					'message'   => Contasimple_CPT::get_message_html( GENERIC_ERROR, null ),
					'icon'      => Contasimple_CPT::get_state_html( GENERIC_ERROR ),
					'buttons'   => Contasimple_CPT::render_action_buttons( $cs_invoice_post, array( 'cs_stop' ) ),
					'date_sync' => date( 'Y-m-d H:i:s' ),
				);
			}
		} else {
			$response = array(
				'error'   => true,
				'message' => __( 'Invalid WooCommerce invoice ID', 'contasimple' ),
			);
		}

		echo wp_json_encode( $response );
		exit();
	}

	/**
	 * Download PDF invoice file from CS
	 *
	 * @since    1.0.0
	 */
	public function cs_pdf() {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		$nonce  = esc_attr( $_REQUEST['_wpnonce'] ); // Input var okay.
		$action = esc_attr( $_REQUEST['action'] ); // Input var okay.

		if ( ! $this->access_allowed( $nonce, $action ) ) {

			$this->logger->logTransactionEnd( 'Access not allowed. Configuration info is missing.' );
			$response = array(
				'redirect' => '',
			);

			echo wp_json_encode( $response );
			exit();

		} else {
			try {
				$id     = (int) $_REQUEST['externalID']; // Input var okay.
				$period = esc_attr( $_REQUEST['period'] ); // Input var okay.
				$number = esc_attr( $_REQUEST['number'] ); // Input var okay.

				$this->cs = new CSService( $this->get_config() );
				$this->cs->getInvoicePDF( $id, $period, $number );

				exit();
			} catch ( \Exception $e ) {
				$response = array(
					'error'   => true,
					'message' => __( 'Invalid WooCommerce invoice ID', 'contasimple' ),
					'summary' => $this->get_summary( null, true ),
				);

				echo wp_json_encode( $response );
				exit();
			}
		}
	}

	/**
	 * Send mail to customer with attached PDF invoice from CS
	 *
	 * @since    1.0.0
	 */
	public function cs_email() {

		$this->logger->logTransactionStart( 'Called: ' . __METHOD__ );

		$nonce  = esc_attr( $_REQUEST['_wpnonce'] ); // Input var okay.
		$action = esc_attr( $_REQUEST['action'] ); // Input var okay.

		if ( ! $this->access_allowed( $nonce, $action ) ) {

			$this->logger->logTransactionEnd( 'Access not allowed.' );
			$response = array(
				'redirect' => admin_url( 'edit.php?post_type=cs_invoice' ),
			);

		} else {
			$cs_invoice_id = (int) $_REQUEST['cs_invoice_id'];

			$contasimple_invoice_sync = new CS_Invoice_Sync( $cs_invoice_id );

            try{
                $order = new WC_Order( $contasimple_invoice_sync->order_id );
            }
            catch(\Exception $ex){
                $order = new WC_Order();
            }

			if ( $cs_invoice_id > 0 & ! empty( $contasimple_invoice_sync ) && ! ( empty( $order ) ) ) {

				$this->maybe_send_email_and_update_status( $order, $contasimple_invoice_sync, true );

				$response = array(
					'success' => true,
					'buttons' => Contasimple_CPT::render_action_buttons( get_post( $cs_invoice_id ), array(
						'cs_view',
						'cs_pdf',
						'cs_email',
					) ),
				);

			} else {
				$response = array(
					'error' => true,
				);
			}
		}

		echo wp_json_encode( $response );
		exit();
	}

	/**
	 * Send the invoice to CS via API
	 *
	 * @param   string $period
	 * @param   array $cs_invoice
	 * @param   int   $externalID
	 *
	 * @return  InvoiceApiModel|\Swagger\Client\Model\AuthenticationResponse
	 * @throws  Exception Containing an error code defined in Service.php.
	 *
	 * @since   1.0.0
	 */
	public function send_invoice( $period, $cs_invoice, $externalID = 0 ) {

		try {
			$this->cs = new CSService( $this->get_config() );
			$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );

			// If the invoice already exists (contains a valid externalID) some minor changes must be made in order to update it instead of creating it
			if ( $externalID && $externalID > 0 ) {
				$response = $this->cs->updateInvoice( $period, $cs_invoice, $externalID );
			} else {
				// If not exceptions thrown and we made it here, it's time to sync the invoice
				$response = $this->cs->createInvoice( $period, $cs_invoice );
			}

			return $response;

		} catch ( \Exception $e ) {
			// The Service class returns the error constant to save on the table as the exception error code
			throw $e;
		}
	}

	/**
	 * Verify that the company stored info in WC is still valid and update it otherwise.
	 *
	 * TODO See if we can use WP transients to do not perform this very time, most likely this info rarely changes.
	 */
	public function maybe_update_cs_company_settings() {

		$this->cs = new CSService( $this->get_config() );
		$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );

		$company = $this->cs->getCurrentCompany();

		if ( $company->getId() != $this->get_config()->getCompanyId() ) {
			$response = $this->cs->selectCompany( $this->cs->getConfig()->getCompanyId() );
			$this->store_config_values( $this->cs->getConfig(), $response->getExpiresIn() + time() );
		} else {
			// Since 1.4.2 > Refresh WC config, maybe user changed its fiscal region, etc.
			$cs_config = $this->cs->getConfig();
			$cs_config->setCurrencySymbol( $company->getExtraInformation()->getCurrencySymbol() );
			$cs_config->setCountryISOCode( $company->getCountry()->getIsoCodeAlpha2() );
			$cs_config->setFiscalRegionCode( $company->getFiscalRegion()->getCode() );
			$cs_config->setVatName( $company->getFiscalRegion()->getVatName() );

			$this->store_config_values( $cs_config );
		}
	}

	/**
	 * Send payment to CS via API
	 *
	 * @param   ContasimpleInvoice $cs_invoice_sync
	 * @param   string             $period
	 * @param   string             $method
	 * @param   float              $amount
	 * @param   null|bool          $reset
	 *
	 * @throws  Exception
	 *
	 * @since   1.0.0
	 */
	public function send_payment( $cs_invoice_sync, $period, $method, $amount, $reset = null ) {

		try {
			$this->cs = new CSService( $this->get_config() );
			$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );

			$CSPayment = array(
				'date'            => $cs_invoice_sync->date_sync,
				'amount'          => $amount,
				'paymentMethodId' => $this->get_payment_id_equivalence( $method )
			);

			$this->cs->assignPayment( $cs_invoice_sync->externalID, $CSPayment, $period );
			$this->logger->log( sprintf( "Payment amount [%1$.2f] synced OK to contasimple invoice with ID [%2d]", $amount, $cs_invoice_sync->externalID ) );

		} catch ( \Exception $e ) {
			$error = $e->getCode();
			$this->logger->log( sprintf( "An error (%1s) occurred trying to send payment amount [%2d] to Contasimple invoice with ID [%3d]", $error, $amount, $cs_invoice_sync->externalID ) );
			throw new \Exception( $e->getMessage(), PAYMENT_SYNC_ERROR, $e );
		}
	}

	/**
	 * Transform an order to an array request object that the CS API can digest
	 *
	 * @param   WC_Order           $order
	 * @param   ContasimpleInvoice $contasimple_invoice_sync
	 * @param   string             $nif
	 * @param   WC_DateTime        $date
	 * @param   null|array         $args
	 *
	 * @return  array|mixed
	 * @throws  Exception
	 *
	 * @since   1.0.0
	 */
	public function convert_order_to_cs_invoice( $order, $contasimple_invoice_sync, $nif, $order_date_completed, $sign, $attempt = 1 ) {

		try {
			if ( empty( $order ) || $order->get_order_number() == null ) {
				throw new \Exception( '', MISSING_ORDER );
			}

			// Contasimple can only handle invoices generated with the same currency as the issuer company's currency set in Contasimple, due to the complexity in currency exchanges
			// Beware, however, that some countries might share the symbol and still they don't use the same currency, ex: USD dollar and Mexican dollar are both $
			if ( $contasimple_invoice_sync->order_currency != $contasimple_invoice_sync->cs_currency ) {
				throw new \Exception( 'Currency symbol of the WC order is ' . $contasimple_invoice_sync->order_currency . ', currency symbol of the CS fiscal region is ' . $contasimple_invoice_sync->cs_currency, INVALID_CURRENCY );
			}

			// Init the API and refresh and re-store tokens if needed
			$this->cs = new CSService( $this->get_config() );
			$this->store_config_values( $this->cs->getConfig(), $this->cs->getConfig()->getExpireTime() );

			// Get target entity data from CS (if customer does not exist, create it, or return aggregated if needed)
			$targetEntity = $this->convert_billing_address_to_cs_customer( $order, $nif );

			// Default mode for entity, used to create an invoice.
			$entityKey = 'target_entity_id';
			$entityVal = $targetEntity->getId();

			// There's a slight difference in the customer entity referencing mode
			// depending on whether it's a new invoice or an update...
			if ( $contasimple_invoice_sync->externalID > 0 && $contasimple_invoice_sync->state === CHANGED ) {
				$entityKey = 'target_entity';
				$entityVal = $targetEntity;
			}

			$order_id     = Contasimple_WC_Backward_Compatibility::get_id( $order );
			$order_status = Contasimple_WC_Backward_Compatibility::get_order_status( $order );

			// Start building the CS invoice entity.
			// The basic info
			$cs_invoice = array(
				$entityKey      => $entityVal,
				'number'        => $contasimple_invoice_sync->number, // Already handled by the API
				'date'          => $order_date_completed,
				'invoice_class' => 700, // TODO: Let's see what has to be here, this value seems to be the default one
				'notes'         => $this->get_invoice_notes( $order_id, $sign, $nif )
			);

			// Set default values just in case some of the if/else does not set the value and the validate amounts detect
			// an undefined index. Needs improvement > refactor needed to minimize spaghetti code.
			$cs_invoice['retentionAmount'] = 0;
			$cs_invoice['retention_percentage'] = 0;

			// Let's decide the invoice type based on availability of invoicing data and issuer/target fiscal regions.
			$cs_invoice = $this->set_operation_type( $order, $cs_invoice );

			// Get candidates for cs invoice lines
			$line_items = $order->get_items( array('line_item', 'shipping', 'fee') );

			// We will keep two instances of the invoice (rounded and not rounded) if the rounded invoice won't sync
			// we might try the 2nd group by VAT strategy with the original un-rounded amounts.
			$cs_invoice_without_rounding = $cs_invoice;

			// Regular orders and full refunds can be iterated in the same fashion, only the amount sign
			// will change (negative) for unitAmount, totalTaxableAmount and vatAmount.
			if ( ! $this->is_partial_refund( $order, $contasimple_invoice_sync->args ) ) {

				// Iterating and extracting data from order products, shipping costs, etc.
				foreach ( $line_items as $item ) {

					$type = Contasimple_WC_Backward_Compatibility::get_order_item_type( $item );

					switch ($type) {
						// It's a shop product
						case 'line_item':

							$concept = Contasimple_WC_Backward_Compatibility::get_order_item_name( $item );
							$unit_amount = $item['line_total'] / $item['qty'];
							$total_taxable_amount = $item['line_total'];
							$quantity = intval( $item['qty'] );

							break;

						// It's a shipping carrier fee
						case 'shipping':

							// Carrier shipping name
							$concept  = __( 'Shipping', 'contasimple' ) . ': ' . Contasimple_WC_Backward_Compatibility::get_order_item_name( $item );
							$unit_amount = Contasimple_WC_Backward_Compatibility::get_order_shipping_item_amount( $item );
							$total_taxable_amount = $unit_amount; // Only 1 item of shipping always!
							$quantity = 1;

							break;

						case 'fee':

							$concept = Contasimple_WC_Backward_Compatibility::get_order_item_name( $item );
							$unit_amount = $item['line_total'];
							$total_taxable_amount = $unit_amount;
							$quantity = 1;

							break;
					}

					$this->add_line_to_cs_invoice( $cs_invoice, $cs_invoice_without_rounding,
						$concept,
						$unit_amount,
						$quantity,
						$total_taxable_amount,
						$item,
						$sign,
						$order
					);
				}

			} else {
				// If it's a refund and the total paid is different than the amount that we get as an argument,
				// it means this is a partial refund, we have to check other entities, account for remaining amounts,
				// etc.
				$this->convert_partial_refund_to_cs_lines( $cs_invoice, $cs_invoice_without_rounding, $contasimple_invoice_sync->args );
			}

			if ( $attempt >= 2 ) {
				$cs_invoice = $this->get_invoice_lines_grouped_by_VAT( $cs_invoice_without_rounding, $order, $sign );
			}

			if ( $attempt == 3 ) {
				if ( count( $cs_invoice['lines'] ) > 1 ) {
					throw new \Exception();
				} else {
					$cs_invoice = $this->get_invoice_one_liner_emergency_sync( $cs_invoice_without_rounding, $order, $sign );
				}
			}

			return $cs_invoice;

		} catch ( \Exception $e ) {
			$this->logger->log( "Exception thrown while gathering PS invoice info to send via API: " . $e->getMessage() );
			throw $e;
		}
	}

	/**
	 * Transforms the Order billing address to a Contasimple customer (targetEntity)
	 *
	 * This is part of the whole Contasimple invoice that we need to send via API.
	 *
	 * @param $order
	 * @param $nif
	 *
	 * @return \Contasimple\Swagger\Client\Model\EntityApiModel|EntityApiModel|null|\Swagger\Client\Model\EntityApiModel
	 */
	public function convert_billing_address_to_cs_customer( $order, $nif ) {

		if ( empty( $nif ) ) {
			// The invoice does not have a valid NIF, we need to get the aggregated customer for tickets.
			$targetEntity = $this->cs->getCustomerAggregated();
		} else {
			$customer_country = Contasimple_WC_Backward_Compatibility::get_billing_country( $order );
			$order_id         = Contasimple_WC_Backward_Compatibility::get_id( $order );
			$country_id       = $this->get_cs_country_id_from_wc_iso( $customer_country );

			if ( ! $this->is_valid_nif( $nif, $customer_country ) ) {
				// If NIF is invalid, do not block the invoice generation, let's send it to the ticket pile (aggregated customer).
				$targetEntity = $this->cs->getCustomerAggregated();
				$nif          = $targetEntity->getNif();
			} else {
				// The customer has a valid NIF, it may exist already in CS or not
				$targetEntity = $this->cs->getCustomerByNIF( $nif );
			}

			if ( empty( $targetEntity ) ) {
				$wc_states = WC()->countries->get_states($customer_country);

				// If it is the first invoice generated for this user, we must create it in CS before sending the invoice, as every
				// CS invoice has its customer embedded as a 'target entity'
				if ( version_compare( WC_VERSION, '2.7', '<' ) ) {
					$billing_company = $order->billing_company;
					$billing_first_name = $order->billing_first_name;
					$billing_last_name = $order->billing_last_name;
					$billing_address_1 = $order->billing_address_1;
					$billing_address_2 = $order->billing_address_2;
					$billing_city = $order->billing_city;
					$billing_postcode = $order->billing_postcode;
					$billing_phone = $order->billing_phone;
					$billing_email = $order->billing_email;
					$billing_state = $order->billing_state;
				} else {
					$billing_company = $order->get_billing_company();
					$billing_first_name = $order->get_billing_first_name();
					$billing_last_name = $order->get_billing_last_name();
					$billing_address_1 = $order->get_billing_address_1();
					$billing_address_2 = $order->get_billing_address_2();
					$billing_city = $order->get_billing_city();
					$billing_postcode = $order->get_billing_postcode();
					$billing_phone = $order->get_billing_phone();
					$billing_email = $order->get_billing_email();
					$billing_state = $order->get_billing_state();
				}

				$cs_customer = array(
					"type"         => "Target",
					"organization" => ! empty( $billing_company ) ? $billing_company : $billing_first_name . " " . $billing_last_name,
					"nif"          => $nif,
					"address"      => $billing_address_1 . " " . $billing_address_2,
					"province"     => empty( $billing_state ) ? "" : html_entity_decode($wc_states[$billing_state]),
					"city"         => $billing_city,
					"country"      => WC()->countries->countries[ $customer_country ],
					"countryID"    => $country_id,
					"postalCode"   => $billing_postcode,
					"phone"        => $billing_phone,
					"fax"          => "",
					"email"        => $billing_email,
					"notes"        => __( 'Synchronized from WooCommerce order #' . $order_id, 'contasimple' ),
					"url"          => "",
					"customField1" => "",
					"customField2" => ""
				);

				$targetEntity = $this->cs->createCustomer( $cs_customer );
			}
		}

		return $targetEntity;
	}

	/**
	 * Handles the particular case of a partial refund.
	 *
	 * @param $cs_invoice The CS invoice entity as we need to inject the changes calculated in this method (by reference).
	 * @param $args The data regarding the WC refund process.
	 *
	 * @return int|number The accumulated decimal error, needed on the outside to decide if we have to switch to the
	 * group by VAT rates calculation strategy or other emergency sync strategies.
	 */
	public function convert_partial_refund_to_cs_lines( &$cs_invoice, $cs_invoice_without_rounding, $args ) {

		if ( is_serialized( $args ) ) {
			$args = unserialize( $args );
		}

		$amount_from_items = 0;

		if ( ! empty( $args['reason'] ) ) {
			$cs_invoice['notes'] .= CS_EOL . __( 'Refund reason: ', 'contasimple' ) . $args['reason'];
		}

		foreach ( $args['line_items'] as $key => $item ) {
			if ( empty( $item['refund_total'] ) ) {
				continue;
			}

			$product_unit_cost = $item['refund_total'] / $item['qty'];
			$product_line_cost = $item['refund_total'];
			$quantity          = intval( $item['qty'] );
			$product_name      = isset( $order_items[ $key ]['name'] ) ? $order_items[ $key ]['name'] : __( 'Other', 'contasimple' );
			$vatAmount         = floatval( $item['refund_tax'][1] );
			$vatPercentage     = floatval( ( $item['refund_tax'][1] * 100 ) / ( $item['refund_total'] ) );

			$line = array(
				'concept'            => $product_name,
				'unitAmount'         => $product_unit_cost,
				'quantity'           => $quantity * -1,
				'vatPercentage'      => $vatPercentage,
				'vatAmount'          => $vatAmount * -1,
				'totalTaxableAmount' => $product_line_cost * -1,
			);

			$line_rounded = $this->round_cs_line_item( $line );

			// Accumulate refunds from items so we can later see if there is some amount without a line associated
			$amount_from_items += $product_line_cost + $vatAmount;

			$cs_invoice['lines'][]                  = $line_rounded;
			$cs_invoice_without_rounding['lines'][] = $line;
		}

		// If there's an orphaned amount, we have to add it as a new line item so that the exact invoice/refund amounts match.
		if ( $amount_from_items < floatval( $args['amount'] ) ) {
			$line = array(
				'concept'            => __( 'Remaining amount refunded', 'contasimple' ),
				'unitAmount'         => ( floatval( $args['amount'] ) - $amount_from_items ),
				'quantity'           => -1,
				'vatPercentage'      => 0, // TODO is this really ok? :/
				'vatAmount'          => 0 * -1,
				'totalTaxableAmount' => ( floatval( $args['amount'] ) - $amount_from_items ) * -1,
			);

			$line_rounded = $this->round_cs_line_item( $line );

			$cs_invoice['lines'][]                  = $line_rounded;
			$cs_invoice_without_rounding['lines'][] = $line;
		}
	}

	public function convert_shipping_to_cs_invoice_line( &$cs_invoice, $order, $sign ) {

		$shipping_items = $order->get_items( 'shipping' );

		if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
			global $wpdb;

			if ( count( $shipping_items ) > 1 ) {
				// WC before v2.2 offers poor support to get tax rates bases on shipping methods as items,
				// so given that almost everybody has dropped support already, we will find a compromise by assuming
				// that only 1 shipping method is used for each order and we will get the total rate + amount used as
				// one CS invoice line.
				// If more than 1 method is detected, throw an error. It should not be a typical scenario.
				throw new \Exception( '', TAXES_PER_LINE_TOO_COMPLEX );
			}

			$shipping = array_shift( $shipping_items );
			$taxes = $order->get_taxes();

			$shipping_taxes_found = 0;
			$shipping_tax = 0;
			$vatPercentage = 0;
			$shipping_total = $order->get_total_shipping();

			foreach ( $taxes as $tax ) {
				if ( isset( $tax['shipping_tax_amount'] ) && $tax['shipping_tax_amount'] > 0 ) {
					$shipping_taxes_found++;
					$shipping_tax = $tax['shipping_tax_amount'];
					$vatPercentage = $wpdb->get_var( $wpdb->prepare( "SELECT tax_rate FROM {$wpdb->prefix}woocommerce_tax_rates WHERE tax_rate_id = %s", $tax['rate_id'] ) );
				}
			}

			// TODO Maybe account for IRPF and R.E in <2.2 since now we know how (see other examples)
			if ( $shipping_tax != $order->get_shipping_tax() && 1 < $shipping_taxes_found ) {
				throw new \Exception( '', TAXES_PER_LINE_TOO_COMPLEX );
			}

			$line = array(
				'concept'            => __( 'Shipping', 'contasimple' ) . ': ' . $shipping['name'],
				'unitAmount'         => $shipping_total * $sign,
				'quantity'           => 1,
				'vatAmount'          => $shipping_tax * $sign,
				'vatPercentage'      => $vatPercentage,
				'totalTaxableAmount' => $shipping_total * $sign,
			);

			$cs_invoice['lines'][] = $line;

		} else {
			// WC 2.2+ offers methods to get shipping concepts more similar to product item lines, it's easier
			// to get the tax rate for each shipping method. Thus, we will support more than one shipping method
			// per order just in case this has some real case use.
			foreach ( $shipping_items as $item ) {
				$shipping_name  = __( 'Shipping', 'contasimple' ) . ': ' . $item->get_name();
				// There's still some differences between up to 2.6 and 3.0+ but not difficult to deal with.
				if ( version_compare( WC_VERSION, '2.7', '<' ) ) {
					$shipping_total = $item['cost'];
					$shipping_tax   = $item['taxes'];
					if( is_serialized( $shipping_tax )) {
						$shipping_tax = maybe_unserialize($shipping_tax);
						$shipping_tax = count($shipping_tax) > 0 ? $shipping_tax[1] : 0;
					}
				} else {
					$shipping_total = $item->get_total();
					$shipping_tax   = $item->get_total_tax();
				}

				// Let's get the VAT for the shipping and handle weird IRPF or R.E. scenarios
				$additionalTaxes = $this->calculate_cs_line_taxes( $item );

				$vatPercentage = $additionalTaxes['vatPercentage'];
				$vatAmount = $additionalTaxes['vatAmount'];
				$rePercentage = $additionalTaxes['rePercentage'];
				$reAmount = $additionalTaxes['reAmount'];

				$line = array(
					'concept'            => $shipping_name,
					'unitAmount'         => $shipping_total * $sign,
					'quantity'           => 1,
					'vatAmount'          => $vatAmount * $sign, //$shipping_tax * $sign,
					'vatPercentage'      => $vatPercentage,
					'totalTaxableAmount' => $shipping_total * $sign,
					'reAmount'           => $reAmount * $sign,
					'rePercentage'       => $rePercentage
				);

				$cs_invoice['lines'][] = $line;
			}
		}
	}

	/**
	 * Rounds the relevant values of a Contasimple invoice line.
	 *
	 * CS forces a line to to be sent with a maximum of CS_MAX_DECIMALS defined decimal numbers (2d as to date).
	 * Away-from-zero rounding is applied.
	 *
	 * @param $line The array containing the original line items, could contain up to 5 digits in decimal values.
	 *
	 * @return An array containing the resulting line items, with their values rounded to CS_MAX_DECIMALS defined
	 * decimal values.
	 */
	public function round_cs_line_item( $line ) {

		$line_rounded = array(
			'concept'            => $line['concept'],
			'unitAmount'         => round( $line['unitAmount'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP ),
			'quantity'           => $line['quantity'],
			'vatPercentage'      => round( $line['vatPercentage'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP ),
			'vatAmount'          => round( $line['vatAmount'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP ),
			'totalTaxableAmount' => round( $line['totalTaxableAmount'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP ),
			'rePercentage'       => round( $line['rePercentage'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP ),
			'reAmount'           => round( $line['reAmount'], CS_MAX_DECIMALS, PHP_ROUND_HALF_UP )
		);

		return $line_rounded;
	}

	/**
	 * Checks if the order refunded amounts and original amount differ.
	 * We need to know this as the process to build our contasimple invoice will be slightly different then if
	 * the refunded amount is the same (full refund).
	 *
	 * @param $order The WC order instance.
	 * @param $args The arguments that come from the order refunded WC hook and contain the amount refunded.
	 *
	 * @return bool True if it is a partial refund, false otherwise.
	 */
	public function is_partial_refund( $order, $args ) {

		if ( is_serialized( $args ) ) {
			$args = unserialize( $args );
		}

		if ( version_compare( WC_VERSION, '2.2', '<' ) ) {
			return ! empty( $args ) && $order->get_total() != $args['amount'];
		} else {
			return ! empty( $args ) && $order->get_total() != $args['amount'];
		}
	}

	/**
	 * Calculate operation type based on issuer and target fiscal information
	 *
	 * @param   WC_Order $order
	 * @param   array    $cs_invoice
	 *
	 * @return  mixed
	 * @throws  Exception
	 *
	 * @since   1.0.0
	 */
	public function set_operation_type( $order, $cs_invoice ) {

		// Note: In this field can be entered a spanish NIF or country equivalent, or also a NIF-IVA aka VIES number.
		$nif       = get_post_meta( $order->get_order_number(), 'NIF', true );
		$countries = new WC_Countries();

		// Without a DNI/VAT, it must be added to the aggregate customer, and that's considered a national operation
		if ( empty( $nif ) ) {
			$cs_invoice['operation_type'] = 'Nacional';
			if ( $order->get_total_tax() == 0 ) {
				$cs_invoice['warning'] = SYNCED_NATIONAL_WITHOUT_TAXES;
			}
		} elseif ( $order->get_total_tax() != 0 ) {
			// If it has any kind of tax, we assume it's VAT, and therefore it must be a national operation as well
			$cs_invoice['operation_type'] = 'Nacional';
			// TODO contemplar inversion del sujeto pasivo y ley digital? :S
		} else {
			// So there's a NIF/VAT NUMBER and there are no taxes... we must determine the reason
			// Gather the data from both parts
			$customerCountry = Contasimple_WC_Backward_Compatibility::get_billing_country( $order );
			$customerState   = Contasimple_WC_Backward_Compatibility::get_billing_state( $order );
			$customerZone    = $countries->get_continent_code_for_country( $customerCountry );

			$issuerCountryISOCode   = $this->get_config()->getCountryISOCode();
			$issuerFiscalRegionCode = $this->get_config()->getFiscalRegionCode();
			$issuerZone = $countries->get_continent_code_for_country( $issuerCountryISOCode );

			$countriesMatch = $issuerCountryISOCode == $customerCountry ? true : false;
			$regionsMatch = self::regions_do_match( $customerState, $issuerFiscalRegionCode );

			// Easiest case to handle: customer or shop are not from Europe or one is but issuer/customer countries do not match -> exportation
			if ( ( $customerZone != "EU" || $issuerZone != "EU") && ! $countriesMatch ) {
				$cs_invoice['operation_type'] = 'Exportacion';
			} elseif ( !$countriesMatch ) {
				// If customer provides a VAT number and the countries don't match, it will be intracommunitary as long
				// as the VAT is valid according to the VIES. Test with FR43542088067 for a success.
				if ( self::vies_web_service_check( $nif ) ) {
					$cs_invoice['operation_type'] = 'Intracumunitaria'; // typo intended
				} else {
					// It would be odd if it's an european operation, there are no taxes and the VIES is invalid. Maybe the customer did supply a non-registered number by error
					// or it could just be a vat exempt operation to another european country?  In any case, it has been decided to mark it as national with a warning message
					$cs_invoice['operation_type'] = 'Nacional';
					$cs_invoice['warning']        = SYNCED_WITH_INVALID_VAT_NUMBER;
				}
			} elseif ( $countriesMatch ) {
				// If it is the same country it should be national, but we must inspect the rules with different tax zones (ex: Canarias, Ceuta & Melilla in Spain)
				if ( $regionsMatch || empty( $customerState ) ) {
					$cs_invoice['operation_type'] = 'Nacional';
					$cs_invoice['warning']        = SYNCED_NATIONAL_WITHOUT_TAXES;
				} else {
					$cs_invoice['operation_type'] = 'Exportacion';
				}
			} else {
				if ( empty( $issuerCountryISOCode ) ) {
					// CS 'Other' countries returns a null value for ISO code, as expected. In that case, we cannot really compare the PS country with the PS issuer country
					// As a last resort, let's check the PS configured shop country instead, and if it is the same as the PS invoice, we will assume a national invoice
					if ( $customerCountry->iso_code == $this->context->country->iso_code ) {
						$cs_invoice['operation_type'] = 'Nacional';
					} else {
						// We will assume exportation?
						$cs_invoice['operation_type'] = 'Exportacion';
					}
				} else {
					throw new \Exception( 'Cannot determine operation type based on taxes applied, customer billing address and company fiscal region address.', OPERATION_TYPE_TOO_COMPLEX );
				}
			}
		}

		return $cs_invoice;
	}

	/**
	 * Reduce invoice lines to lines grouped by VAT percentage
	 *
	 * This reduces the rounding error, invoke when we detect a threshold error that is not tolerable.
	 *
	 * @param   $cs_invoice
	 * @param   $order
	 *
	 * @return  mixed
	 * @throws  Exception
	 *
	 * @since   1.0.0
	 */
	public function get_invoice_lines_grouped_by_VAT( $cs_invoice, $order, $sign ) {

		try {
			$currency  = $order->get_currency();
			$line_num  = 1;
			$new_lines = array();

			// We will keep the original context of the lines inside the CS notes field, so that the user can get some insight about what it was
			foreach ( $cs_invoice['lines'] as $line ) {
				$cs_invoice['notes'] .= CS_EOL . CS_EOL . sprintf( __( 'Concept %1$d: %2$s', 'contasimple' ), $line_num, $line['concept'] ) . CS_EOL;
				$cs_invoice['notes'] .= sprintf( __( '%1$1.6f%2$s x %3$d qty, %4$1.3f%% Tax (%5$1.6f%2$s), TOTAL: %6$1.6f%2$s', 'contasimple' ), $line['unitAmount'], $currency, $line['quantity'], $line['vatPercentage'], $line['vatAmount'], $line['vatAmount'] + $line['unitAmount'] * $line['quantity'] ) . CS_EOL;
				$cs_invoice['notes'] .= CS_EOL;

				$line_num ++;
			}

			$cs_invoice['notes'] = __( 'The concepts of the invoice do not reflect the individual products of the order because the taxable amount charged for one or more products (the unit cost excluding taxes and having applied any discount) is not representable with two decimals, which is why the order had to be synchronized by grouping the totals by type of tax. If you want your eCommerce order invoices to be synchronized by keeping the original product lines, the price assigned to the products from your shop must be an amount that, after deducting taxes and discounts, can be represented exactly with two decimal digits and without the need to round off the amount.', 'contasimple' )
				. CS_EOL . CS_EOL
				. __( 'These are the details of the original invoice lines:', 'contasimple' )
				. CS_EOL . CS_EOL
				. $cs_invoice['notes'];

			// Condense N lines as concepts into M lines grouped by VAT type
			foreach ( $cs_invoice['lines'] as $line ) {
				$key = $line['vatPercentage'];

				$new_lines[ $key ] = array(
					'concept'            => sprintf( __( 'Sum of the products with %1$d%% VAT', 'contasimple' ), $line['vatPercentage'] ),
					'unitAmount'         => ( isset( $new_lines[ $key ]['unitAmount'] ) ? $new_lines[ $key ]['unitAmount'] : 0 ) + $line['unitAmount'] * $line['quantity'],
					'quantity'           => 1 * $sign,
					'vatPercentage'      => $line['vatPercentage'],
					'vatAmount'          => ( isset( $new_lines[ $key ]['vatAmount'] ) ? $new_lines[ $key ]['vatAmount'] : 0 ) + $line['vatAmount'],
					'totalTaxableAmount' => ( isset( $new_lines[ $key ]['totalTaxableAmount'] ) ? $new_lines[ $key ]['totalTaxableAmount'] : 0 ) + $line['totalTaxableAmount'],
					'rePercentage'       => $line['rePercentage'],
					'reAmount'           => ( isset( $new_lines[ $key ]['reAmount'] ) ? $new_lines[ $key ]['reAmount'] : 0 ) + $line['reAmount']
				);
			}

			// Once grouped, now do the final rounding at the very last moment.
			foreach ( $new_lines as &$new_line ) {
				$new_line = $this->round_cs_line_item( $new_line );
			}

			// Reset keys or otherwise the API would fail
			$cs_invoice['lines'] = array_values( $new_lines );

			return $cs_invoice;
		} catch ( \Exception $e ) {
			$this->logger->log( "Exception thrown while generating a simplified invoice to avoid rounding issues: " . $e->getMessage() );
			throw $e;
		}
	}

	public function get_invoice_one_liner_emergency_sync( $cs_invoice, $order, $sign ) {
		try {

			// Check if a line contains no VAT amount nor prices (a zero line, might happen).
			// In this case this is not a deal breaker, we can remove it and still condense the invoice.
			foreach ( $cs_invoice['lines'] as $index => $line ) {
				if ( $line['totalTaxableAmount'] == 0 && $line['vatAmount'] == 0 ) {
					unset( $cs_invoice['lines'][$index] );
				}
			}

			// Now there must only remain one line with some kind of VAT rate.
			if ( count( $cs_invoice['lines'] ) !== 1 ) {
				throw new \Exception( 'Previously grouped by VAT invoice returning more than one line. This invoice has several different taxes applied and thus cannot be condensed into a one line invoice.', TOTAL_AMOUNT_INVALID );
			}

			// Amounts will come directly from the totals calculated by WC.
			$orderTotalAmountWithoutVAT = $order->get_total() - $order->get_total_tax();
			$orderTotalVAT = +$order->get_total_tax();

			// The only remaining VAT rate must be the correct one.
			$tax = array_shift( $cs_invoice['lines'] );

			if ( !empty( $tax ) && is_array( $tax ) ) {
				$orderVatPercentage = $tax['vatPercentage'];
			} else {
				// In the event that there was an issue with the CS calculated line vat, do the manual calculation.
				$orderVatPercentage = round(( $order->get_total() / ($order->get_total() - $order->get_total_tax()) - 1 ) * 100, 1, PHP_ROUND_HALF_UP);
			}

			// Reset lines.
			unset( $cs_invoice['lines'] );

			// ... and inject our custom one.
			$cs_invoice['lines'][] = array(
				'concept' => __('Order total', 'contasimple'),
				'quantity' => 1 * $sign,
				'unitAmount' =>  $orderTotalAmountWithoutVAT,
				'vatAmount' => $orderTotalVAT * $sign,
				'vatPercentage' =>  $orderVatPercentage,
				'totalTaxableAmount' => $orderTotalAmountWithoutVAT
			);

			return $cs_invoice;

		} catch (\Exception $e) {
			$this->logger->log( "Exception thrown while generating an emergency sync invoice to avoid rounding issues: " . $e->getMessage() );
			throw $e;
		}
	}

	/**
	 * Get the CS equivalent payment method ID from WC selected method
	 *
	 * @param $method
	 *
	 * @return mixed
	 *
	 * @since   1.0.0
	 */
	public function get_payment_id_equivalence( $method ) {
		// Get the equivalences configured in the wizard
		$paymentEquivalences = $this->cs->getConfig()->getPaymentEquivalences();

		// It should be set, but if it is not, use the default one
		if ( ! array_key_exists( $method, $paymentEquivalences ) ) {
			$method = "default";
		}

		// Return an integer that the API can consume
		return $paymentEquivalences[ $method ];
	}

	/**
	 * Store company ID from the selected company via AJAX
	 *
	 * An APIkey will return one or more companies associated to a CS account. The user picks one and the wizard
	 * invokes this method to store the selection using WP Options.
	 *
	 * @since   1.0.0
	 */
	public function cs_select_company() {

		$result = array(
			'message' => __( 'Your company preferences have been updated', 'contasimple' )
		);

		try {
			$this->check_ajax_wizard_allowed();

			$companyId = (int) $_REQUEST['company'];
			if ( ! empty( $companyId ) ) {
				$registrationNumber = ''; // Configuration::get('PS_SHOP_DETAILS');
				$this->cs           = new CSService( $this->config );
				$companiesList      = $this->cs->getCompanies();
				foreach ( $companiesList as $company ) {
					if ( $company->getId() == $companyId ) {
						if ( $company->getRequiresConfiguration() ) {
							$result = array(
								'error'   => true,
								'message' => __( 'Your Contasimple account is not fully configured. Please login into Contasimple.com to complete the configuration process.', 'contasimple' )
							);
						} else {
							if ( ! empty( $company->getExtraInformation() ) ) {
								if ( ! $this->registration_number_matches( $registrationNumber, $company->getExtraInformation()->getEntity()->getNif() ) ) {
									$result = array(
										'error'                 => true,
										'updateCompanyInfoStep' => true,
										'message'               => __( 'Your account needs additional configuration. Please wait...', 'contasimple' )
									);
								} else {
									$response = $this->cs->selectCompany( $companyId );

									// Changing company implies new refreshToken, save it now, otherwise if the code
									// breaks for some reason, reconnecting will trigger an 'invalid refresh token'
									// error.
									$this->store_config_values( $this->cs->getConfig(), $response->getExpiresIn() + time() );

									// Get payments data and pass it down to the last step
									$responsePaymentMethods = $this->cs->getPaymentMethods();
									foreach ( $responsePaymentMethods as $paymentMethod ) {
										$paymentMethods[] = array(
											'id_option' => $paymentMethod->getId(),
											'name'      => $paymentMethod->getName(),
										);
									}
									$result['paymentMethods'] = $paymentMethods;

									// $this->cs->getConfig()->setCurrencyISOCode( $company->getCountry()->getCurrency()->getShortName() );
									$this->cs->getConfig()->setCurrencySymbol( $company->getExtraInformation()->getCurrencySymbol() );
									$this->cs->getConfig()->setCountryISOCode( $company->getCountry()->getIsoCodeAlpha2() );
									$this->cs->getConfig()->setFiscalRegionCode( $company->getFiscalRegion()->getCode() );
									$this->cs->getConfig()->setVatName( $company->getFiscalRegion()->getVatName() );

									$this->store_config_values( $this->cs->getConfig(), $response->getExpiresIn() + time() );
								}
							} else {
								$result = array(
									'error'   => true,
									'message' => __( 'Your Contasimple account is not fully configured. Please login into Contasimple.com to complete the configuration process.', 'contasimple' )
								);
							}
						}

						break;
					}
				}
			}
		} catch ( Exception $e ) {
			if ( $e->getCode() == API_ERROR ) {
				// If we get an API error, pass down this message to the end user as most likely this error is not
				// controlled and the error might be of interest.
				$result = array(
					'error'   => true,
					'message' => $e->getMessage()
				);
			} elseif ( $e->getCode() == INSUFFICIENT_RIGHTS_ERROR ) {
				// In the very special case that the user is picking a company which does not have enough rights
				// to use with the plugin, display a custom error, as the error returned by the API is too generic.
				$result = array(
					'error'   => true,
					'message' => __( 'User permissions in the selected company do not support synchronization with WooCommerce, please use user credentials with the appropriate permissions.', 'contasimple' )
				);
			} else {
				// If it's not an API error, just display a generic error regarding the company completion process.
				$result = array(
					'error'   => true,
					'message' => __( 'Your company selection could not be set. Please try again, and if the problem persists, contact us for technical support.', 'contasimple' )
				);
			}
		}

		echo wp_json_encode( $result );
		wp_die();
	}

	/**
	 * Store CS payment methods equivalences in WC via AJAX
	 *
	 * Invoked by the config wizard
	 *
	 * @since   1.0.0
	 */
	public function cs_select_payment_methods() {

		$result = array(
			'message' => __( 'Your payment preferences have been updated', 'contasimple' ) . ". " . __( 'Please, wait...', 'contasimple' )
		);

		try {
			$this->check_ajax_wizard_allowed();

			$paymentMethods = array();
			parse_str( $_REQUEST['data'], $paymentMethods );

			if ( ! empty( $paymentMethods ) ) {
				$this->config->setPaymentEquivalences( $paymentMethods );
				$this->store_config_values( $this->config );
			}
		} catch ( Exception $e ) {
			$result = array(
				'error'   => true,
				'message' => __( $e->getMessage(), 'contasimple' )
			);
		}

		echo wp_json_encode( $result );
		wp_die();
	}

	/**
	 * Check access to a certain section is granted
	 *
	 * Uses WP nounces for validation and only allows access also if company info available by running the config wizard
	 *
	 * @param $nonce
	 * @param $action
	 *
	 * @return bool
	 */
	public function access_allowed( $nonce, $action ) {

		return function_exists( 'curl_version' ) && class_exists( 'Contasimple' ) && self::properly_configured() && wp_verify_nonce( $nonce, $action );
	}

	/**
	 * Check that AJAX calls during settings wizard are allowed
	 *
	 * We reuse the WC settings nonce
	 */
	public function check_ajax_wizard_allowed( ) {

		if ( ! check_ajax_referer( 'woocommerce-settings', null, false ) ) {

			$this->logger->log( 'Referrer WP nonce check failed. AJAX call not allowed.' );

			$result = array(
				'error'   => true,
				'message' => __( 'Access denied. Make sure you are still logged into WordPress.', 'contasimple' ),
			);

			echo wp_json_encode( $result );
			wp_die();
		}
	}

	/**
	 * Check that all data needed from running the config wizard exists
	 *
	 * We need to have an API key for a company to make all the other sections work.
	 *
	 * @return  bool      If config data is available
	 * @throws  Exception If something went work with the checks
	 *
	 * @since   1.0.0
	 */
	public static function properly_configured() {
		$config = get_option( 'contasimple_settings_account' );

		if ( empty( $config ) || $config instanceof __PHP_Incomplete_Class ) {
			$config = new CSConfig();
		}

		try {
			return ! (
				empty( $config->getAccessToken() )
				|| empty( $config->getRefreshToken() )
				|| empty( $config->getApiKey( 'apikey' ) )
				|| empty( $config->getCompanyId() )
				|| empty( $config->getExpireTime() )
				|| empty( $config->getPaymentEquivalences() )
			);
		} catch ( \Exception $e ) {
			throw $e;
		}
	}

	/**
	 * Delete current user settings (forces the request of needed user info)
	 *
	 * Deletes all stored contasimple variables from ps_configuration table.
	 * The configuration wizard will have to be run again to reconfigure the module.
	 *
	 * @since   1.0.0
	 */
	public function force_login() {

		delete_option( 'contasimple_settings_account' );
		$this->config = new CSConfig();

		return true;
	}

	/**
	 * Check if two NIFs match
	 *
	 * Used to check if WC shop owner and CS selected company fiscal ID do match.
	 *
	 * @param $numberWP
	 * @param $numberCS
	 *
	 * @return bool
	 *
	 * @deprecated  It seems we are no longer doing this check since it complicates too much initial configuration.
	 */
	public static function registration_number_matches( $numberWP, $numberCS ) {
		$numberWP = strtolower( preg_replace( "/[^[:alnum:]]/u", "", $numberWP ) );
		$numberCS = strtolower( preg_replace( "/[^[:alnum:]]/u", "", $numberCS ) );

		//return $numberWP === $numberCS;
		return true; // TODO Decide what to do, remove this validation or implement somehow in WP¿
	}

	/**
	 * Check if there exist logs for the selected date
	 *
	 * @since   1.0.0
	 */
	public function check_for_log() {
		try {
			$this->check_ajax_wizard_allowed();

			$result = false;
			$date   = esc_attr( $_REQUEST['date'] );

			$logs_dir  = plugin_dir_path( __DIR__ ) . "logs" . DIRECTORY_SEPARATOR;
			$file_name = "contasimple_" . date( 'd-m-Y', strtotime( str_replace( '/', '-', $date ) ) ) . ".log";
			$full_path = $logs_dir . $file_name;

			if ( file_exists( $full_path ) ) {
				$result = true;
			} else {
			}
		} catch ( \Exception $e ) {
		}

		echo wp_json_encode( $result );
		wp_die();
	}

	/**
	 * Download log file for the selected date
	 *
	 * @param $date
	 *
	 * @since   1.0.0
	 */
	public function download_cs_log( $date ) {
		$logs_dir  = plugin_dir_path( __DIR__ ) . "logs" . DIRECTORY_SEPARATOR;
		$file_name = "contasimple_" . date( 'd-m-Y', strtotime( str_replace( '/', '-', $date ) ) ) . ".log";
		$full_path = $logs_dir . $file_name;

		header( 'Content-Type: application/text' );
		header( 'Content-Disposition: attachment; filename=' . $file_name );
		header( 'Pragma: no-cache' );

		readfile( $full_path );

		exit();
	}

	/**
	 * Attach a PDF file to our invoice email class
	 *
	 * @param   $attachments
	 * @param   null $email_id
	 * @param   null $order
	 * @param   null $cs_invoice
	 *
	 * @return  array
	 *
	 * @since   1.0.0
	 */
	public function attach_cs_invoice( $attachments, $email_id = null, $order = null, $cs_invoice = null ) {

		// Only for our custom CS Invoice email type.
		if ( $email_id === 'wc_cs_invoice_generated' && ! empty( $cs_invoice ) ) {

			$id     = $cs_invoice->externalID;
			$date   = date( 'Y-m-d', strtotime( $cs_invoice->date_sync ) ); //$order->get_date_completed()->date( 'Y-m-d' );
			$period = Contasimple_CPT::get_period_from_date( $date );
			$number = $cs_invoice->number;

			$this->cs      = new CSService( $this->get_config() );
			$pdf_as_string = $this->cs->getInvoicePDFasString( $id, $period, $number );

			$upload_dir = wp_upload_dir();

			if ( ! empty( $upload_dir['basedir'] ) ) {
				$plugin_dirname = $upload_dir['basedir'] . '/cs_pdf_invoices';
				if ( ! file_exists( $plugin_dirname ) ) {
					wp_mkdir_p( $plugin_dirname );
				}
			}

			$file = mb_ereg_replace( "([^\w\s\d\-_~,;\[\]\(\).])", '', $cs_invoice->number );
			$file = mb_ereg_replace( "([\.]{2,})", '', $file );

			$file_path = $plugin_dirname . DIRECTORY_SEPARATOR . $file . '.pdf';

			file_put_contents( $file_path, $pdf_as_string );

			$attachments[] = $file_path;
		}

		return $attachments;
	}

	/**
	 * Allow screen options in invoices page
	 *
	 * @since   1.0.0
	 */
	public function add_options() {

		$option = 'per_page';
		$args = array(
			'label' => __('Orders per page', 'contasimple'),
			'default' => 10,
			'option' => 'orders_per_page'
		);

		add_screen_option( $option, $args );

		// Adds help screen, enhancement?
		/*
		get_current_screen()->add_help_tab( array(
			'id'		=> 'custom',
			'title'		=> __('Custom'),
			'content'	=>
				'<p>' . __('This screen provides access to all of your posts. You can customize the display of this screen to suit your workflow.') . '</p>'
		) );
		*/
	}

	/**
	 * Add our desired options to options screen
	 *
	 * Like how many per page
	 *
	 * @since   1.0.0
	 */
	public function set_screen_options() {

		add_filter('set-screen-option', array(&$this, 'set_option'), 10, 3);
	}

	public function set_option($status, $option, $value) {

		return $value;
	}

	/**
	 * Add a menu entry and define page URL for previous orders importation
	 *
	 * @since   1.0.0
	 */
	public function contasimple_add_previous_woocommerce_orders() {

		// Register the hidden sub-menu.
		$hook = add_submenu_page(
			'woocommerce' // Use the parent slug as usual.
			, __( 'Import previous orders', 'contasimple' )
			, ''
			, 'manage_options'
			, 'create_cs_invoices'
			, array(&$this, 'display_create_cs_invoices' )
		);

		add_action( "load-$hook", array (&$this, 'add_options') );
	}

	/**
	 * Make the new entry menu look nice
	 *
	 * @param   $submenu_file
	 *
	 * @return  string
	 *
	 * @since   1.0.0
	 */
	function contasimple_admin_submenu_filter( $submenu_file ) {

		global $plugin_page;

		$hidden_submenus = array(
			'create_cs_invoices' => true,
		);

		// Select another submenu item to highlight (optional).
		if ( $plugin_page && isset( $hidden_submenus[ $plugin_page ] ) ) {
			$submenu_file = 'edit.php?post_type=shop_order';
		}

		// Hide the submenu.
		foreach ( $hidden_submenus as $submenu => $unused ) {
			remove_submenu_page( 'woocommerce', $submenu );
		}

		return $submenu_file;
	}

	/**
	 * Create a page to import invoices from previous orders
	 *
	 * This makes use of the WP List Table class.
	 *
	 * @since   1.0.0
	 */
	public function display_create_cs_invoices() {

		$myListTable = new Contasimple_WP_List_Table_WC_Orders();

		$doaction = $myListTable->current_action();

		// If there's a form action, process it!
		if ( $doaction ) {
			switch ( $doaction ) {
				case 'create':
					$this->logger->logTransactionStart( 'Called: ' . __METHOD__ . ' with bulk invoice create action.' );

					$order_ids = array_reverse( apply_filters( 'request', $_REQUEST['order_ids'] ) );
					$total = count( $order_ids );
					$count = 0;

					// If there are any checked rows, get their IDs and iterate and add the invoice
					if ( $total > 0 ) {
						$unique_series_prefix = substr(md5(microtime()),rand(0,26),4) . "_";
						foreach ( $order_ids as $order_id ) {
							// If order has refunds see if we can add it or better skip.
							$order = Contasimple_WC_Backward_Compatibility::get_order_from_id( $order_id );
							$order_invoices = CS_Invoice_Sync::get_invoices_from_order( $order_id );
							$has_refunds = Contasimple_WC_Backward_Compatibility::get_invoice_sign_from_order( $order ) < 0;

							if ($this->invoice_can_be_added( $order_invoices, $has_refunds ) ) {
								$contasimple_invoice_sync = CS_Invoice_Sync::create_empty( $order_id );
								if ( ! empty( $contasimple_invoice_sync ) ) {
									if ( self::properly_configured() ) {
										$contasimple_invoice_sync->companyID = $this->config->getCompanyId();
										$contasimple_invoice_sync->mask      = $unique_series_prefix . $this->get_mask_from_order( $order_id );
										$contasimple_invoice_sync->save();
										$count++;
										$this->logger->log( 'Added CS invoice to queue from order id ' . $order_id . ' with mask ' . $contasimple_invoice_sync->mask );
									} else {
										$this->logger->log( 'Trying to add previous invoices to the queue but configuration missing? Should not be here.' );
									}
								} else {
									$this->logger->log( 'Error creating an entry into the database for the previous order.' );
								}
							} else {
								$this->logger->log( 'Order has refunds and no original invoice found, skipping refund.' );
							}
						}

						// We must redirect the user to this same page without the post data, otherwise we could potentially duplicate invoices!
						$new_url = add_query_arg(
							array (
								'count' => $count,
								'total' => $total,
							),
							menu_page_url( 'create_cs_invoices', false )
						);

						wp_redirect( $new_url, 303 );
					}
					break;
			}
		} else {
			$count = filter_input( INPUT_GET, 'count' );
			$total = filter_input( INPUT_GET, 'total' );;

			// Only if we added these two query args previously, it means we already added invoices and we want to inform the end user
			if ( !empty( $count ) && ! empty ( $total ) ) {
				if ($total == $count) {
					$result_mesasge = new Contasimple_Notice( sprintf( __( 'Invoices successfully added to the queue.', 'contasimple' ), $count, $total ), 'success' );
				} else {
					$result_mesasge = new Contasimple_Notice( sprintf( __( '%1$d out of %2$d invoices added. Some errors where found.', 'contasimple' ), $count, $total ), 'error' );
				}

				$result_mesasge->render();
			}

			// In any case, build the table list and show remaining orders
			$myListTable->render_before_table();
			$myListTable->prepare_items();
			$myListTable->display();
			$myListTable->render_after_table();
		}
	}

	/**
	 * Check that the order total and vat amounts and the amounts gathered by the plugin match before sending
	 *
	 * @param $order
	 * @param $cs_invoice
	 *
	 * @throws Exception In case that the amounts do not match
	 *
	 * @since   1.0.0
	 */
	public function validate_invoice_amounts( $order, &$cs_invoice, $sign = 1 ) {

		$totalWithoutVat = 0;
		$totalVat = 0;
		$totalReAmount = 0;
		$totalIrpfAmount = 0;
		$vatRate = 0;

		if ( !empty( $cs_invoice['lines'] ) ) {
			foreach ($cs_invoice['lines'] as $line) {
				$totalWithoutVat += $line['totalTaxableAmount'] * $sign;
				$totalVat += $line['vatAmount'] * $sign;
				if(!empty($line['reAmount'])){
				    $totalReAmount += $line['reAmount'] * $sign;
                }
                if(!empty($line['vatPercentage'])) {
	                $vatRate = $line['vatPercentage'];
                }
			}
		}

		$order_subtotal = ( $order->get_total() - $order->get_total_tax() );
        $order_tax_total = +$order->get_total_tax();
		$order_expected_retention = ( $order_subtotal * $cs_invoice['retention_percentage'] / 100 ) * $sign;

		// The epsilon is very small number used because of PHP issues with float comparison due to internal
		// representation and is equivalent to comparing with zero.
		$epsilon = 0.0001;

		if ( !empty( $cs_invoice['retentionAmount'] ) ) {
			$totalIrpfAmount = $cs_invoice['retentionAmount'];

			// The IRPF amount that we reported to CS should be equal to the original subtotal multiplied by the IRPF
			// rate, if not, then some of the line items did not have IRPF and we cannot let it sync.
			if ( abs ( $order_expected_retention -  $totalIrpfAmount ) < $epsilon  ) {
				$this->logger->log( 'Order IRPF values in WC match the calculated by CS, OK!' );
			} else {
				throw new Exception( 'Order IRPF (' . $totalIrpfAmount . ') does not match CS IRPF total amounts (' . $order_expected_retention . '), KO!', IRPF_PER_LINE_NOT_ALLOWED );
			}
		}

		$totalTaxes =  $totalVat + $totalReAmount - $totalIrpfAmount;

		$totalAmountDifference = round( $order_subtotal - $totalWithoutVat, CS_MAX_DECIMALS, PHP_ROUND_HALF_UP );
		$totalVATAmountDifference = round( $order_tax_total - $totalTaxes, CS_MAX_DECIMALS, PHP_ROUND_HALF_UP );

		if ( abs( $totalAmountDifference ) < $epsilon && abs( $order_tax_total - $totalTaxes ) < $epsilon ) {
			$this->logger->log( 'Order total and vat amounts match those calculated by CS, OK!' );
			return 1;

		} else if ( abs( $totalAmountDifference ) <= CS_ROUNDING_THRESHOLD && abs( $totalVATAmountDifference ) <= CS_ROUNDING_THRESHOLD  ) {
			// Amounts differ due to rounding issues but up to a certain tolerable threshold,
			// so add the remaining cents as a new invoice line so that total amounts will match in both platforms.
			$minimum = 0.01;

			if ( abs( $totalAmountDifference ) >= $minimum ) {
				$this->logger->log( 'Order total computed by CS differs slightly from WC, sync OK but with WARNING.' );
				$line = array(
					'concept'            => __( 'Rounding adjustment for total price without VAT', 'contasimple' ),
					'unitAmount'         => $order_subtotal - $totalWithoutVat,
					'quantity'           => 1 * $sign, // handle refund via qty * -1
					'totalTaxableAmount' => $order_subtotal - $totalWithoutVat,
					'vatPercentage'      => 0,
					'vatAmount'          => 0,
					'rePercentage'       => 0,
					'reAmount'           => 0
				);

				$cs_invoice['lines'][] = $this->round_cs_line_item( $line );
			}

			if ( abs( $totalVATAmountDifference ) >= $minimum ) {
				$this->logger->log( 'Total VAT computed by CS differs slightly from WC, sync OK but with WARNING.' );
				$line = array(
					'concept'            => __( 'Rounding adjustment for VAT amount', 'contasimple' ),
					'unitAmount'         => 0,
					'quantity'           => 1 * $sign, // handle refund via qty * -1
					'totalTaxableAmount' => 0,
					'vatPercentage'      => $vatRate,
					'vatAmount'          => ( $order_tax_total - $totalTaxes ) * $sign,
					'rePercentage'       => 0,
					'reAmount'           => 0
				);

				$cs_invoice['lines'][] = $this->round_cs_line_item( $line );
			}

			return 0;
		} else {
			$this->logger->log( 'WC Order total (' . $order_subtotal . ') and VAT amounts (' . $order_tax_total
			                    . ') DO NOT match those calculated by CS ('
			                    . $totalWithoutVat . ' and ' . $totalTaxes . ' respectively), KO!');
			return -1;
		}
	}

	/**
	 * Check customer NIF validity based on country defined REGEX
	 *
	 * We retrieve the countries list from the CS API
	 *
	 * @param $nif
	 * @param $countryIsoCode
	 *
	 * @return bool
	 *
	 * @since   1.0.0
	 */
	public function is_valid_nif( $nif, $countryIsoCode ) {

		$countries = $this->get_cached_cs_countries();

		foreach ( $countries as $country ) {
			if ( $country->getIsoCodeAlpha2() == $countryIsoCode ) {
				$pattern = '/' . $country->getNifValidationRegex() . '/';

				if ( ! preg_match( $pattern, $nif ) ) {
					return false;
				}
			}
		}

		return true;
	}

	/**
	 * Get the customer country in CS based on WC country ID
	 *
	 * @param $customer_country
	 *
	 * @return int If not matching found, will return 999 (other for CS)
	 *
	 * @since   1.0.0
	 */
	public function get_cs_country_id_from_wc_iso( $customer_country ) {

		$customer_country_in_cs = 999;
		$countries = $this->get_cached_cs_countries();

		foreach ( $countries as $country ) {
			if ( $country->getIsoCodeAlpha2() == $customer_country ) {
				$customer_country_in_cs = $country->getId();
				break;
			}
		}

		return $customer_country_in_cs;
	}

	/**
	 * Retrieve countries from CS API
	 *
	 * We use transients to cache them temporally since this should not change frequently.
	 *
	 * @return  mixed|\Swagger\Client\Model\CountryApiModel[]
	 *
	 * @since   1.0.0
	 */
	public function get_cached_cs_countries() {
		// Read about WP Transient API.
		$countries = get_transient( 'country_API_results' );

		// If not already set, fetch countries from the API and save the transient.
		if ( false === $countries ) {
			try {
				$contasimple_admin = new Contasimple_Admin();
				$contasimple_admin->init_service();
				$countries = $contasimple_admin->get_service()->getCountries( 'true' );
				set_transient( 'country_API_results', $countries, COUNTRIES_EXPIRATION_IN_SECONDS );
			} catch ( \Exception $e ) {
				$this->logger->logTransactionEnd( 'Exception thrown while trying to load CS countries: ' . $e->getMessage() );
				$countries = array();
			}
		} else {
			// We avoid calling the API for subsequent NIF checks until the transient expires.
			$countries = get_transient( 'country_API_results' );
		}

		return $countries;
	}


	/**
	 * Return an array with VAT and RE values and checks that additional VAT rates are either RE or IRPF
	 *
	 * TODO The internals of this method should be reviewed, probably there's a better way to handle this.
	 * Also it only works for Spain.
	 *
	 * @param $order
	 * @param $taxes
	 *
	 * @return An array with the VAT and R.E. amounts and percentages found (0 for each not found).
	 *
	 * @throws Exception If none of the additional vat rates are either RE or IRPF
	 */
	public function calculate_cs_line_taxes ( $item, $order ) {

		$type = Contasimple_WC_Backward_Compatibility::get_order_item_type( $item );

		switch( $type ) {

			case 'line_item':
				$product    = Contasimple_WC_Backward_Compatibility::get_product_from_order_item( $item );
				$item_taxes = Contasimple_WC_Backward_Compatibility::get_order_item_taxes( $item, $product );
				break;

			case 'shipping':
				$item_taxes = Contasimple_WC_Backward_Compatibility::get_order_shipping_item_taxes( $item, $order );
				break;

			case 'fee':
				$item_taxes = Contasimple_WC_Backward_Compatibility::get_order_fee_item_taxes( $item ); //$item->get_taxes()['total'];
				break;
		}

		$cs_line_taxes = array(
			'vatAmount'            => 0,
			'vatPercentage'        => 0,
			'reAmount'             => 0,
			'rePercentage'         => 0,
			'retention_percentage' => 0,
			'retentionAmount'      => 0,

		);

		if ( ! isset( $item_taxes ) || count( $item_taxes ) == 0 ) {
			return $cs_line_taxes;
		}

		$is_req = false;
		$is_irpf = false;

		$deleted_taxes = false;

		foreach ( $item_taxes as $tax_id => $tax_amount ) { // https://docs.woocommerce.com/wc-apidocs/source-class-WC_Tax.html#651-673
			// Skip weird cases of empty value taxes but forming part of the tax array.
			if ( empty( $tax_amount ) ) {
				unset( $item_taxes[$tax_id] );
				continue;
			}

			$tax_name = WC_Tax::get_rate_label($tax_id);
			$tax_name_lower = strtolower( $tax_name );
			$tax_rate = Contasimple_WC_Backward_Compatibility::get_tax_rate( $tax_id );

			// Should not happen, but what if tax deleted?
			if ($tax_amount > 0 && $tax_rate == 0) {
				$deleted_taxes = true;
				$this->logger->log( "Tax rate from id is 0 but vat amount detected, maybe tax was deleted?" );
				$this->logger->log( "Trying to calculate tax rate from amounts with formula..." );
				$tax_rate = Contasimple_WC_Backward_Compatibility::calculate_tax_rate_from_amounts( $item, $tax_amount );
				$this->logger->log( "Calculated VAT rate: " . $tax_rate );
			}

			// If any of these values are found in the tax line title, we assume it's a valid R.E tax
			// TODO Simplify this or find a better strategy!
			if ( strpos( $tax_name_lower, 'equivalencia' ) === 0 ||
			     strpos( $tax_name_lower, 're' ) === 0 ||
			     strpos( $tax_name_lower, 'r.e' ) === 0 ||
			     strpos( $tax_name_lower, '0.5' ) === 0 ||
			     strpos( $tax_name_lower, '0,5' ) === 0 ||
			     strpos( $tax_name_lower, '1.4' ) === 0 ||
			     strpos( $tax_name_lower, '1,4' ) === 0 ||
			     strpos( $tax_name_lower, '1.75' ) === 0 ||
			     strpos( $tax_name_lower, '1,75' ) === 0 ||
			     strpos( $tax_name_lower, '5.2' ) === 0 ||
			     strpos( $tax_name_lower, '5,2' ) === 0 ||
			     $tax_rate == 5.2 ||
			     $tax_rate == 1.4 ||
			     $tax_rate == 1.75
			) {
				$cs_line_taxes['reAmount']     = $tax_amount;
				$cs_line_taxes['rePercentage'] = $tax_rate;

				$is_req = true;
			}

			else if ( strpos( $tax_name_lower,'irpf' ) === 0 || strpos( $tax_name_lower,'i.r.p.f' ) === 0) {
				// IRPF is not part of a line in CS, but we return it to build the total sum for the invoice.
				$is_irpf = true;
				$cs_line_taxes['retention_percentage'] = $tax_rate * -1;
				$cs_line_taxes['retentionAmount'] = $tax_amount * -1;
			}

			else {
				$cs_line_taxes['vatPercentage'] = $tax_rate;
				$cs_line_taxes['vatAmount'] = $tax_amount;
			}
		}

		if (!(
			0 === count( $item_taxes ) ||
			1 === count( $item_taxes ) && ! $is_req ||             // Only VAT or IRPF but not R.E is OK
			2 === count( $item_taxes ) && $is_req && ! $is_irpf || // Only VAT and R.E is OK
			2 === count( $item_taxes ) && $is_irpf && ! $is_req || // Only VAT and IRPF is OK
			3 === count( $item_taxes ) && $is_req && $is_irpf      // Only VAT, R.E and IRPF is OK
		))
		{
			$this->logger->log("Item taxes: " . count( $item_taxes) .
			                   ", is_req: " . $is_req == true ? "true" : "false" .
			                   ", is_irpf: " . $is_irpf == true ? "true" : "false" .
			                   ", taxes: " . json_encode( $item_taxes ) ) .
							   ", deleted taxes: " . $deleted_taxes ? "true" : "false";

			if ( $deleted_taxes ) {
				throw new \Exception( '', DELETED_TAXES );
			} else {
				throw new \Exception( '', TAXES_PER_LINE_TOO_COMPLEX );
			}
		}

		return $cs_line_taxes;
	}

	public function add_line_to_cs_invoice(	&$cs_invoice, &$cs_invoice_without_rounding,
		$concept,
		$unit_amount,
		$quantity,
		$total_taxable_amount,
		$item,
		$sign,
		$order
	) {

		$line_taxes = $this->calculate_cs_line_taxes( $item, $order );

		// For now, keep a line with all it's decimal details.
		$line = array(
			'concept'            => $concept,
			'unitAmount'         => $unit_amount,
			'quantity'           => $quantity * $sign, // handle refund via qty * -1
			'totalTaxableAmount' => $total_taxable_amount * $sign,
			'vatPercentage'      => $line_taxes['vatPercentage'],
			'vatAmount'          => $line_taxes['vatAmount'] * $sign,
			'rePercentage'       => $line_taxes['rePercentage'],
			'reAmount'           => $line_taxes['reAmount']
		);

		// Round the line > this is what we will send to CS.
		$line_rounded = $this->round_cs_line_item( $line );

		$cs_invoice['lines'][] = $line_rounded;
		$cs_invoice_without_rounding['lines'][] = $line;

		// IRPF (if any) is not accounted at the line level in CS, but the total order.
		$cs_invoice['retentionAmount'] += $line_taxes['retentionAmount'] * $sign;

		// If a line has no IRPF percentage, let's keep the previous value, maybe this line has a 0 price and should
		// not make the invoice to break.
		if ( $line_taxes['retention_percentage'] > 0 ) {
			$cs_invoice['retention_percentage'] = $line_taxes['retention_percentage'];
		}
	}

	/**
	 * Returns the difference between the rounded line and the original line
	 *
	 * @deprecated as of 1.4.2 > We are using the triple strategy now.
	 *
	 * @param $original_line
	 * @param $line_rounded
	 *
	 * @return number
	 */
	public function get_accumulated_rounding_error( $original_line, $line_rounded ) {

		$line_error = abs(
			    ( $original_line['unitAmount'] * $original_line['quantity'] + $original_line['vatAmount'] )
			  - ( $line_rounded['unitAmount'] * $line_rounded['quantity'] + $line_rounded['vatAmount'] )
		);

		return $line_error;
	}

    /**
     * Gets the IRPF total amount for the current order
     *
     * @param $order
     *
     * @return mixed
     */
    public function get_irpf_total( $order ) {

        $req = $order->get_data()['tax_lines'];

        $vatIrpfAmount = 0;

        foreach ($req as $tx_line){
            $lbl_tax = $tx_line['label'];
            $lbl_lower = strtolower($lbl_tax);
            $rate_id = $tx_line['rate_id'];


            if ( strpos($lbl_lower,'irpf') === 0 || strpos($lbl_lower,'i.r.p.f') === 0){
                $vatIrpfAmount += $tx_line['tax_total'];
            }
        }


        return $vatIrpfAmount;
    }

    public function get_irpf_shipping( $shipping_taxes ) {

	    $vatIrpfAmount = 0;

	    if ( isset( $shipping_taxes ) ) {

		    foreach ( array_shift( $shipping_taxes ) as $key => $tx_line ) {
			    $rate_id = $key;
			    $lbl_tax = WC_Tax::_get_tax_rate( $rate_id )['tax_rate_name'];
			    $lbl_lower = strtolower($lbl_tax);

			    if ( strpos($lbl_lower,'irpf') === 0 || strpos($lbl_lower,'i.r.p.f') === 0){
				    $vatIrpfAmount += $tx_line;
			    }
		    }
	    }

	    return $vatIrpfAmount;
    }

    /**
     * Gets the IRPF percentage for the current order
     *
     * @param $order
     *
     * @return mixed
     */
    public function get_irpf_percentage($order) {

	    $vatIrpfPercentage = 0;

        $req = $order->get_data()['tax_lines'];

        foreach ($req as $tx_line){
            $lbl_tax = $tx_line['label'];
            $lbl_lower = strtolower($lbl_tax);
            $rate_id = $tx_line['rate_id'];


            if ( strpos($lbl_lower,'irpf') === 0 || strpos($lbl_lower,'i.r.p.f') === 0){
                $vatIrpfPercentage = WC_Tax::_get_tax_rate( $rate_id )['tax_rate'];
            }
        }
        return round( $vatIrpfPercentage, CS_MAX_DECIMALS, PHP_ROUND_HALF_UP );
    }

    public static function send_error_sync_mail( $contasimple_invoice, $sync_result )
    {
        // We need to force the mailer init so that it can register the hook 'woocommerce_' for our custom invoice email, it does not load by default.
        WC()->mailer();

        $res = apply_filters( 'woocommerce_cs_invoice_error_sync', $contasimple_invoice, $sync_result, 2 );

        return $res;
    }

	/**
	 * Checks that a new invoice can be added to the sync pool.
	 *
	 * When the order status (completed or refunded) hook is triggered, we need to create an invoice and send it to CS.
	 * However, there is a problem that the admin could keep changing the order status forever and that would trigger
	 * a new invoice generation, causing duplicates.
	 *
	 * This method will check if the WP hook is eligible to add the new invoice or
	 * if we should just resume the queue to sync already marked invoices, as any change to the original invoice will
	 * be automatically accounted for later in the sync process (attention: this only works with invoices not synced OK yet).
	 *
	 * This might change in the future, but for the time being we decided that an invoice can be synced if:
	 * 1) Is the first one with a positive value (a regular invoice)
	 * 2) We are coming from the refund hook, but already exists an invoice (otherwise it does not make sense to create
	 *    an order slip if we have not officially invoiced the order, this might happen if status 'Refunded' is picked
	 *    before even picking the 'Completed' status.
	 *
	 * @param $order_invoices  An array of (already existing) Contasimple_Invoice_Sync entities.
	 * @param $isRefund        Indicates if we are coming from a refund (true) or an invoice (false)
	 * @since 1.4.0
	 *
	 * @return bool
	 */
    private function invoice_can_be_added( $order_invoices, $isRefund ) {

	    $canBeAdded = count( $order_invoices ) == 0 && ! $isRefund ||
	                  count( $order_invoices ) == 1 && (float) $order_invoices[0]->amount > 0 && $isRefund;

	    return $canBeAdded;
    }

	/**
	 * Gets the notes text that will be added to the Contasimple invoice.
	 *
	 * Use this to trespass any details that will be lost from the WooCommerce context and might be of interest to
	 * the shop owner.
	 *
	 * @param $order_id
	 * @param $sign
	 * @param $nif
	 *
	 * @return string
	 */
	public function get_invoice_notes( $order_id, $sign, $nif ) {

		if ( $sign < 0 ) {
			$invoice_type = __( 'Order refund', 'contasimple');
		} else {
			$invoice_type = __( 'Order', 'contasimple');
		}

		$message = sprintf(
			__( '%1$s synchronized from WooCommerce (order #%2$d)', 'contasimple' ),
			$invoice_type,
			$order_id
		);

		$order = Contasimple_WC_Backward_Compatibility::get_order_from_id( $order_id );
		$customer_country = Contasimple_WC_Backward_Compatibility::get_billing_country( $order );

		$nif_invalid = empty( $nif ) || !$this->is_valid_nif( $nif, $customer_country );

		if ( $nif_invalid ) {
			$customer = Contasimple_WC_Backward_Compatibility::get_customer_full_name( $order );

			if ( !empty( $customer ) || !empty( $nif ) ) {
				$message .= CS_EOL . sprintf(
					__( 'Customer name and NIF (if available): %1$s %2$s', 'contasimple' ),
					$customer,
					empty( $nif ) ? '' : '(' . $nif . ')'
				) . CS_EOL;
			}
		}

		return $message;
	}

	/**
	 * Gets the Contasimple company identifier name.
	 *
	 * @return string A string value. Example: 'NIF' for Spain.
	 *                Returns an empty string if the country is not found.
	 */
	public function get_company_identifier_name() {

		$config_company_identifier_name = '';

		// Look for the fiscal region ID number name (ie: NIF for Spain)
		foreach ( $this->get_cached_cs_countries() as $country ) {
			if ( $country->getIsoCodeAlpha2() == $this->get_config()->getCountryISOCode() ) {
				$regions = $country->getFiscalRegions();
				if ( count( $regions ) > 0 ) {
					$config_company_identifier_name = strval($regions[0]['company_identifier_name']);
				}
				break;
			}
		}

		return $config_company_identifier_name;
		
	}

	public static function regions_do_match( $key, $value ) {

		$equivalencyArray = array(
			'ES-TF' => 'es_Canarias', // 'España - Canarias'
			'ES-GC' => 'es_Canarias', // 'España - Canarias'
			'ES-CE' => 'es_CeutaYMelilla', // 'España - Ceuta y Melilla'
			'ES-ML' => 'es_CeutaYMelilla', // 'España - Ceuta y Melilla'
		);

		return isset($equivalencyArray[$key]) && $equivalencyArray[$key] == $value;
	}

	/**
	 * Check VAT number against VIES
	 *
	 * Inspired by PrestaShop European VAT Module (see \vatnumber module)
	 *
	 * @param $vat_number
	 * @return bool         VAT number is valid or not
	 * @throws Exception    If web request times out
	 */
	public static function vies_web_service_check( $vat_number ) {
		if ( empty( $vat_number ) ) {
			return false;
		}

		$vat_number = str_replace( ' ', '', $vat_number );
		$prefix = substr( $vat_number, 0, 2 );

		if ( array_search( $prefix, self::get_prefix_intracom_vat() ) === false ) {
			return false;
		}

		$vat = substr( $vat_number, 2 );
		$url = 'http://ec.europa.eu/taxation_customs/vies/viesquer.do?ms='. urlencode( $prefix )
		       .'&iso='.urlencode( $prefix )
		       .'&vat='.urlencode( $vat );
		@ini_set( 'default_socket_timeout', 2 );

		for ( $i = 0; $i < 3; $i++ ) {
			if ( $page_res = file_get_contents( $url ) ) {
				if ( preg_match( '/invalid VAT number/i', $page_res ) ) {
					@ini_restore( 'default_socket_timeout' );
					return false;
				} else if ( preg_match( '/valid VAT number/i', $page_res ) ) {
					@ini_restore( 'default_socket_timeout' );
					return true;
				} else {
					++$i;
				}
			} else {
				sleep( 1 );
			}
		}
		@ini_restore( 'default_socket_timeout' );
		throw new \Exception( '', VIES_TIMEOUT );
	}

	/**
	 * A list of valid European VAT country prefixes, also from vatnumber module
	 * @return array
	 */
	public static function get_prefix_intracom_vat() {
		$intracom_array = array(
			'AT' => 'AT',
			//Austria
			'BE' => 'BE',
			//Belgium
			'DK' => 'DK',
			//Denmark
			'FI' => 'FI',
			//Finland
			'FR' => 'FR',
			//France
			'FX' => 'FR',
			//France métropolitaine
			'DE' => 'DE',
			//Germany
			'GR' => 'EL',
			//Greece
			'IE' => 'IE',
			//Irland
			'IT' => 'IT',
			//Italy
			'LU' => 'LU',
			//Luxembourg
			'NL' => 'NL',
			//Netherlands
			'PT' => 'PT',
			//Portugal
			'ES' => 'ES',
			//Spain
			'SE' => 'SE',
			//Sweden
			'GB' => 'GB',
			//United Kingdom
			'CY' => 'CY',
			//Cyprus
			'EE' => 'EE',
			//Estonia
			'HU' => 'HU',
			//Hungary
			'LV' => 'LV',
			//Latvia
			'LT' => 'LT',
			//Lithuania
			'MT' => 'MT',
			//Malta
			'PL' => 'PL',
			//Poland
			'SK' => 'SK',
			//Slovakia
			'CZ' => 'CZ',
			//Czech Republic
			'SI' => 'SI',
			//Slovenia
			'RO' => 'RO',
			//Romania
			'BG' => 'BG',
			//Bulgaria
			'HR' => 'HR',
			//Croatia
		);

		return $intracom_array;
	}
}
