<?php
/**
 * Contasimple Integration for WooCommerce
 *
 * @link       http://www.contasimple.com
 * @since      1.0.0
 *
 * @package    contasimple
 * @category   contasimple/includes
 * @author     Contasimple S.L. <soporte@contasimple.com>
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}

if ( ! class_exists( 'WC_Integration_Contasimple' ) ) :
	/**
	 * Contasimple Integration Class
	 *
	 * @since    1.0.0
	 */
	class WC_Integration_Contasimple extends WC_Integration {

		/**
		 * A reference to the Contasimple_Admin class that defines a few methods needed for checks.
		 *
		 * @since    1.0.0
		 *
		 * @var $contasimple_admin
		 */
		private $contasimple_admin;

		/**
		 * Init and hook-in the integration.
		 *
		 * @since    1.0.0
		 */
		public function __construct() {

			global $pagenow;

			$this->id                 = 'integration-contasimple';
			$this->method_title       = __( 'Contasimple', 'contasimple' );
			$this->method_description = __( 'Welcome to Contasimple.', 'contasimple' );

			$this->contasimple_admin = new Contasimple_Admin();

			// Custom handling of post actions.
			$this->post_process();

			// Only try to add the integration forms if we are under the desired section since we try to connect with the API
			// to validate the access and would be very unefficient to do this on very constructor call.
			if ( ( 'admin.php' === $pagenow && ('wc-settings' === $_GET['page'] || 'create_cs_invoices' === $_GET['page'] ) )
			     || ( 'edit.php' === $pagenow && isset( $_GET['post_type'] ) && 'cs_invoice' === $_GET['post_type'] ) ) {

				// Load the settings, depending on if configuration is needed or not.
				if ( ! $this->contasimple_admin->properly_configured() ) {
					$this->init_form_wizard_fields();
					if ( ( 'edit.php' === $pagenow && isset( $_GET['post_type'] ) && 'cs_invoice' === $_GET['post_type'] ) ) {
						wp_redirect( admin_url() );
						exit();
					}
				} else {
					try {
						$this->contasimple_admin->init_service(); // TODO is this always necessary?
						$company = $this->contasimple_admin->get_service()->getCurrentCompany();

						if ( ! empty( $company ) ) {
							$this->init_after_wizard_form( $company );
						} else {
							throw new \Exception( "Could not load CS company data.", CANNOT_READ_COMPANY_DATA );
						}
					} catch ( \Exception $e ) {
						$this->handle_exception( $e );
					}
				}
			}
		}

		/**
		 * Initialize integration settings form fields.
		 *
		 * These are only needed in the wizard, will let us pick payment methods to tie to Contasimple counterparts.
		 *
		 * @since    1.0.0
		 */
		public function init_form_wizard_fields() {

			$enabled_payment_gateways = array();
			$payment_gateways_obj     = new WC_Payment_Gateways();

			foreach ( $payment_gateways_obj->payment_gateways() as $gateway ) {
				$enabled_payment_gateways[] = array(
					'name'        => $gateway->id,
					'displayName' => $gateway->title,
				);
			}

			// Add an additional one that will be picked if something goes wrong or if new methods are added in the future.
			$enabled_payment_gateways[] = array(
				'name'        => 'default',
				'displayName' => __( 'Default method', 'contasimple' ),
			);

			$this->form_fields = array(
				'wizard'    => array(
					'title'             => __( 'Configuration Wizard', 'contasimple' ),
					'custom_attributes' => array(
						'paymentModulesList' => $enabled_payment_gateways,
					),
					'type'              => 'modal',
				),
                'enable_logs'           => array(
                    'title'       => __( 'Enable Logs', 'contasimple' ),
                    'type'        => 'checkbox',
                    'label'       => __( 'Enable logging', 'contasimple' ),
                    'default'     => 'yes',
                    'description' => __( 'Contasimple error logs are needed in order to be able to help you in case you might have issues with the plugin. It is recommended to keep them enabled.', 'contasimple' ),
                ),
                'log_files_date'        => array(
                    'title' => __( 'Log files', 'contasimple' ),
                    'type'  => 'text',
                    'class' => 'datepicker',
                    'label' => __( 'Log files', 'contasimple' ),
                ),
                'log_files_button'      => array(
                    'title'       => __( 'Download', 'contasimple' ),
                    'type'        => 'button',
                    'id'          => 'logDownloadButton',
                    'class'       => 'button-primary',
                    'css'         => 'margin-top:8px',
                    'description' => __( 'Select the date for which you want to download the log file and click the \'Download\' button. This file can be sent to Contasimple\'s customer service to help you trooubleshoot any issues with the module .', 'contasimple' ) . __( 'Then, please contact us first by filling the form in the \'contact area\' in our website and we will provide you with further instructions about how to send us the logs.', 'contasimple' ),
                ),
				'version'   => array(
					'title' => __( 'Module version', 'contasimple' ),
					'type'  => 'label',
					'css'   => '',
					'text'  => Contasimple::$version,
				),
			);
		}

		/**
		 * Initialize integration settings form fields.
		 *
		 * These will be needed after the wizard is completed.
		 */
		public function init_form_fields( $company = null ) {

			if ( ! empty( $company ) ) {
				$nif     = '';
				$address = '';

				if ( ! empty( $company->getExtraInformation() ) ) {
					$address = $company->getExtraInformation()->getEntity()->getAddress();
					$nif = $company->getExtraInformation()->getEntity()->getNif();
				}

				$this->form_fields = array(
					'summary'               => array(
						'title'    => __( 'Summary', 'contasimple' ),
						'email'    => $this->contasimple_admin->get_config()->getUsername(),
						'company'  => $company->getOrganizationName(),
						// 'currency' => $company->getCountry()->getCurrency()->getShortName(),
						'nif'      => $nif,
						'address'  => $address,
						'type'     => 'summary',
						'desc_tip' => false,
					),
					'configuration-section' => array(
						'title' => __( 'Configuration', 'contasimple' ),
						'type'  => 'custom_heading',
						'icon'  => 'dashicons-admin-generic',
					),
					'invoices_mask'         => array(
						'title'       => __( 'Autonumber formatting for invoices', 'contasimple' ),
						'type'        => 'text',
						'default'     => __( 'WC-INV-AAAA-######', 'contasimple' ),
						'description' => __( 'Specify the format for the numbering sequence of your invoices.', 'contasimple' ) . ' ' . __( 'Example: WC-INV-AAAA-###### will turn into a sequence like WC-INV-2017-000001', 'contasimple' ),
					),
					'receipts_mask'         => array(
						'title'       => __( 'Autonumber formatting for receipts', 'contasimple' ),
						'type'        => 'text',
						'default'     => __( 'WC-TICK-AAAA-######', 'contasimple' ),
						'description' => __( 'Specify the format for the numbering sequence of your receipts.', 'contasimple' ) . ' ' . __( 'Example: WC-TICK-AAAA-###### will turn into a sequence like WC-TICK-2017-000001', 'contasimple' ),
					),
					'show_warnings'         => array(
						'title'       => __( 'Show warnings', 'contasimple' ),
						'type'        => 'checkbox',
						'label'       => __( 'Enable warnings', 'contasimple' ),
						'default'     => 'no',
						'description' => __( 'Detect if invoices that do not follow Contasimple fiscal region tax rules have been synced.', 'contasimple' ),
					),
					'enable_logs'           => array(
						'title'       => __( 'Enable Logs', 'contasimple' ),
						'type'        => 'checkbox',
						'label'       => __( 'Enable logging', 'contasimple' ),
						'default'     => 'no',
						'description' => __( 'Contasimple error logs are needed in order to be able to help you in case you might have issues with the plugin. It is recommended to keep them enabled.', 'contasimple' ),
					),
					'log_files_date'        => array(
						'title' => __( 'Log files', 'contasimple' ),
						'type'  => 'text',
						'class' => 'datepicker',
						'label' => __( 'Log files', 'contasimple' ),
					),
					'log_files_button'      => array(
						'title'       => __( 'Download', 'contasimple' ),
						'type'        => 'button',
						'id'          => 'logDownloadButton',
						'class'       => 'button-primary',
						'css'         => 'margin-top:8px',
						'description' => __( 'Select the date for which you want to download the log file and click the \'Download\' button. This file can be sent to Contasimple\'s customer service to help you trooubleshoot any issues with the module .', 'contasimple' ) . __( 'Then, please contact us first by filling the form in the \'contact area\' in our website and we will provide you with further instructions about how to send us the logs.', 'contasimple' ),
					),
					'version'               => array(
						'title' => __( 'Module version', 'contasimple' ),
						'type'  => 'label',
						'css'   => '',
						'text'  => Contasimple::$version,
					),
				);

				// Refunds came with WC 2.2
				if ( ! version_compare( WC_VERSION, '2.2', '<' ) ) {
					$refunds_mask_section = array(
						'title'       => __( 'Autonumber formatting for refunds', 'contasimple' ),
						'type'        => 'text',
						'default'     => __( 'WC-NOTE-AAAA-######', 'contasimple' ),
						'description' => __( 'Specify the format for the numbering sequence of your refunds/credit notes.', 'contasimple' ) . ' ' . __( 'Example: WC-NOTE-AAAA-###### will turn into a sequence like WC-NOTE-2017-000001', 'contasimple' ),
					);

					$this->form_fields = array_slice( $this->form_fields, 0, 3, true) + array('refunds_mask' => $refunds_mask_section ) + array_slice($this->form_fields, 3, count($this->form_fields) - 1, true) ;
				}
			}
		}

		/**
		 * Generate Button HTML.
		 *
		 * @access public
		 * @param mixed $key the key.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_button_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'class'             => 'button-secondary',
				'id'                => $field,
				'css'               => '',
				'custom_attributes' => array(),
				'desc_tip'          => false,
				'title'             => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();
			?>

			<form id="form-cs-download-log" method="post" class="defaultForm form-horizontal" enctype="multipart/form-data">
				<input type="hidden" id="_wpnonce" name="_wpnonce" value="<?php echo wp_create_nonce('woocommerce-settings'); ?>">
				<input type="hidden" name="_wp_http_referer" value="<?php echo admin_url( 'admin.php?page=wc-settings&tab=integration&section=integration-contasimple'); ?>">
				<tr valign="top">
					<td></td>
					<td style="padding-top:0px">
						<fieldset>
							<legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span></legend>
							<?php
								echo $this->get_description_html( $data ); // XSS: sanitization okay.
							?>
							<button class="<?php echo esc_attr( $data['class'] ); ?>" type="submit" name="<?php echo empty( $data['id'] ) ? esc_attr( $field ) : esc_attr( $data['id'] ); ?>" id="<?php echo empty( $data['id'] ) ? esc_attr( $field ) : esc_attr( $data['id'] ); ?>" style="<?php echo esc_attr( $data['css'] ); ?>" <?php echo esc_html( $this->get_custom_attribute_html( $data ) ); ?>><?php echo wp_kses_post( $data['title'] ); ?></button>
						</fieldset>
					</td>
				</tr>
			</form>

			<?php
			return ob_get_clean();
		}

		/**
		 * Generate Label HTML.
		 *
		 * @access public
		 * @param mixed $key the kjey.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_label_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'class'    => '',
				'css'      => '',
				'desc_tip' => false,
				'title'    => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();
			?>

			<tr valign="top">
				<td>
					<label class="<?php echo esc_attr( $data['class'] ); ?>" for="<?php echo esc_attr( $field ); ?>" id="<?php echo esc_attr( $field ); ?>" style="<?php echo esc_attr( $data['css'] ); ?>" <?php echo esc_html( $this->get_custom_attribute_html( $data ) ); ?>><?php echo wp_kses_post( $data['title'] ); ?></label>
				</td>
				<td>
					<fieldset>
					<label class="<?php echo esc_attr( $data['class'] ); ?>" for="<?php echo esc_attr( $field ); ?>" id="<?php echo esc_attr( $field ); ?>" style="<?php echo esc_attr( $data['css'] ); ?>" <?php echo esc_html( $this->get_custom_attribute_html( $data ) ); ?>><?php echo wp_kses_post( $data['text'] ); ?></label>
				</td>
			</tr>

			<?php
			return ob_get_clean();
		}

		/**
		 * Generate Modal HTML (for Bootstrap).
		 *
		 * @access public
		 * @param mixed $key the key.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_modal_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'class'    => '',
				'css'      => '',
				'desc_tip' => false,
				'title'    => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();

			require plugin_dir_path( dirname( __FILE__ ) ) . 'admin/partials/contasimple-wizard-display.php';

			return ob_get_clean();
		}

		/**
		 * Generate a custom HTML control with company data.
		 *
		 * @access public
		 * @param mixed $key the key.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_summary_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'class'    => '',
				'css'      => '',
				'company'  => '',
				'desc_tip' => false,
				'title'    => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();

			require plugin_dir_path( dirname( __FILE__ ) ) . 'admin/partials/contasimple-company-summary.php';

			return ob_get_clean();
		}

		/**
		 * Generate HTML links.
		 *
		 * Mainly to give access to the documentation PDF in different languages in config page.
		 *
		 * @access public
		 * @param mixed $key the key.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_links_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'class'             => '',
				'css'               => '',
				'custom_attributes' => array(),
				'desc_tip'          => false,
				'title'             => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();

			require plugin_dir_path( dirname( __FILE__ ) ) . 'admin/partials/contasimple-links.php';

			return ob_get_clean();
		}

		/**
		 * Generate HTML for custom heading and description in config page.
		 *
		 * Just to show a somewhat customized heading, the WP form helpers are too rigid.
		 *
		 * @access public
		 * @param mixed $key the key.
		 * @param mixed $data the data.
		 *
		 * @since 1.0.0
		 * @return string
		 */
		public function generate_custom_heading_html( $key, $data ) {

			$field    = $this->plugin_id . $this->id . '_' . $key;
			$defaults = array(
				'id'    => $field,
				'icon'  => '',
				'title' => '',
			);

			$data = wp_parse_args( $data, $defaults );

			ob_start();

			require plugin_dir_path( dirname( __FILE__ ) ) . 'admin/partials/contasimple-custom-heading.php';

			return ob_get_clean();
		}

		/**
		 * Handle post actions to current page.
		 *
		 * The WC_Settings helpers only post data for the fields definded with its API, we have to deal specifically with our own custom fields and buttons.
		 **/
		protected function post_process() {

			// If we are doing a custom post we have to inspect for custom post fields added by AJAX calls or other POST gimmick.
			if ( isset( $_REQUEST['confirm-unlink'] ) // Input var okay.
			     && true === (bool) $_REQUEST['confirm-unlink'] // Input var okay.
			     && wp_verify_nonce( $_REQUEST['_wpnonce'], 'woocommerce-settings' ) // Input var okay.
			) {
				$this->contasimple_admin->force_login();
			}

			if ( isset( $_REQUEST['confirm-reset'] ) // Input var okay.
			     && true === (bool) $_REQUEST['confirm-reset'] // Input var okay.
			     && wp_verify_nonce( $_REQUEST['_wpnonce'], 'woocommerce-settings' ) // Input var okay.
			) {
				$this->contasimple_admin->force_login();

				$invoices_posts = get_posts( array( 'post_type' => 'cs_invoice', 'numberposts' => -1 ) );

				foreach ( $invoices_posts as $invoice_post ) {
					// Delete's each post.
					wp_delete_post( $invoice_post->ID, true );
					// Set to False if you want to send them to Trash.
				}

				// TODO notice?
			}

			if ( isset( $_REQUEST['refresh'] ) // Input var okay.
			     && true === (bool) $_REQUEST['refresh'] // Input var okay.
			     && wp_verify_nonce( $_REQUEST['_wpnonce'], 'woocommerce-settings' ) // Input var okay.
			) {
				$this->contasimple_admin->init_service();

				$company = $this->contasimple_admin->get_service()->getCurrentCompany();
				$cs_config = $this->contasimple_admin->get_config();

				$cs_config->setCurrencySymbol( $company->getExtraInformation()->getCurrencySymbol() );
				$cs_config->setCountryISOCode( $company->getCountry()->getIsoCodeAlpha2() );
				$cs_config->setFiscalRegionCode( $company->getFiscalRegion()->getCode() );
				$cs_config->setVatName( $company->getFiscalRegion()->getVatName() );

				$this->contasimple_admin->store_config_values( $cs_config );

				new Contasimple_Notice( __( 'Contasimple company settings updated', 'contasimple' ), 'success' );
			}

			// Download custom log file for selected date.
			if ( isset( $_REQUEST['log-date'] )
			     && ! empty( $_REQUEST['log-date'] )
				 && wp_verify_nonce( $_REQUEST['_wpnonce'], 'woocommerce-settings' )
			) {
				$this->contasimple_admin->download_cs_log( sanitize_text_field( wp_unslash( $_REQUEST['log-date'] ) ) ); // Input var okay.
			}

			if ( isset( $_REQUEST['sync'] )
				&& 'resume' === sanitize_text_field( wp_unslash( $_REQUEST['sync'] ) )
				&& wp_verify_nonce( $_REQUEST['_wpnonce'], 'resume' )
			) {
				add_action( 'init', array( $this->contasimple_admin, 'resume_queue_manually' ), 10, 0 );
			}
		}

		/**
		 * Handle exceptions thrown during the integration init.
		 * @param Exception $e
		 *
		 * @since 1.5
		 */
		protected function handle_exception( \Exception $e ) {

			// First of all, log the error for further investigation.
			if ( !empty( $this->contasimple_admin ) && !empty( $this->contasimple_admin->getLogger() ) ) {
				$this->contasimple_admin->getLogger()->log( "Exception thrown while trying to load the plugin during " . get_class( $this ) . " _construct : Code: " . $e->getCode(). ", Message: " . $e->getMessage() );
			}

			switch( $e->getCode() ) {
				case AUTH_DENIED:
					$this->contasimple_admin->force_login();
					$this->init_form_wizard_fields();
					new Contasimple_Notice( __( 'Contasimple login failed. Either your company status or access credentials have changed. Please re-run the configuration wizard to start working with Contasimple again.', 'contasimple' ), 'error' );

					break;

				case HOST_UNREACHABLE:
					new Contasimple_Notice( __( 'Could not connect with Contasimple. Please make sure your server has Internet access and try again and if the problem persists, please contact us.', 'contasimple' ), 'error' );
					$this->init_after_wizard_form( new \Contasimple\Swagger\Client\Model\CompanyApiModel() );
					break;

				case API_ERROR:
					if ( !empty( $e->getMessage() ) ) {
						new Contasimple_Notice( $e->getMessage(), 'error' );
					} else {
						new Contasimple_Notice( __( 'An unknown error occurred trying to fetch your Contasimple account data, please contact us.', 'contasimple' ), 'error' );
					}
					new Contasimple_Notice( $e->getMessage(), 'error' );
					$this->init_after_wizard_form( new \Contasimple\Swagger\Client\Model\CompanyApiModel() );
					break;

				case CANNOT_READ_COMPANY_DATA:
				default:
					new Contasimple_Notice( __( 'An unknown error occurred trying to fetch your Contasimple account data, please contact us.', 'contasimple' ), 'error' );
					$this->init_after_wizard_form( new \Contasimple\Swagger\Client\Model\CompanyApiModel() );
			}
		}

		/**
		 * Loads the config form with company data plus other CS options like invoicing mask, download log, etc.
		 * @param $company \Contasimple\Swagger\Client\Model\CompanyApiModel
		 *        An object containing info about the company.
		 *
		 * @since 1.5
		 */
		protected function init_after_wizard_form( $company = null ) {

			$this->init_form_fields( $company );
			$this->init_settings();

			add_action( 'woocommerce_update_options_integration_' . $this->id, array( $this, 'process_admin_options' ) );
		}
	}

endif;
