<?php
/**
 * Facade class for the Contasimple API
 *
 * This file contains a facade class which makes it easier to communicate with the CS Rest APÎ.
 *
 * @link       https://wordpress.org/plugins/contasimple/
 * @since      1.0.0
 *
 * @package    contasimple
 * @subpackage contasimple/includes/common
 * @author     Contasimple S.L. <soporte@contasimple.com>
 */

use \Contasimple\Swagger\Client\ApiClient as ContasimpleApiClient;
use \Contasimple\Swagger\Client\ApiException as ContasimpleApiException;
use \Contasimple\Swagger\Client\API\AuthenticationApi as ContasimpleAuthenticationApi;
use \Contasimple\Swagger\Client\API\MeCompaniesApi as ContasimpleMeCompaniesApi;
use \Contasimple\Swagger\Client\API\EntitiesCustomersApi as ContasimpleEntitiesCustomersApi;
use \Contasimple\Swagger\Client\API\EntitiesCustomersAggregatedApi as ContasimpleEntitiesCustomersAggregatedApi;
use \Contasimple\Swagger\Client\API\AccountingInvoicesIssuedApi as ContasimpleAccountingInvoicesIssuedApi;
use \Contasimple\Swagger\Client\API\AccountingInvoicesIssuedPaymentsApi as ContasimpleAccountingInvoicesIssuedPaymentsApi;
use \Contasimple\Swagger\Client\API\ConfigurationPaymentMethodsApi as ContasimpleConfigurationPaymentMethodsApi;
use \Contasimple\Swagger\Client\API\ConfigurationCommonApi as ContasimpleConfigurationCommonApi;
use \Contasimple\Swagger\Client\Model\AuthenticationResponse;
use \Contasimple\Swagger\Client\Model\ConfigureCompanyApiModel;
use \Contasimple\Swagger\Client\Model\CreateIssuedInvoiceApiModel;
use \Contasimple\Swagger\Client\Model\UpdateIssuedInvoiceApiModel;

const CS_EOL = "\r\n";
const CS_ROUNDING_THRESHOLD = 0.01;

// We ignore API textual error messages and we define custom constants for every scenario.
// This way the consuming PHP modules for each webapp can use the constants to define their own error messages
// based on the application reporting style, multi-language needed or not, etc.

const NOT_SYNC = 0;
const PENDING = 1;
const SYNC_OK = 2;
const CHANGED = 3;
const PAUSED = 4;
// 5 to 9 reserved for future potential 'non-error' states
const SYNCED_WITH_INVALID_VAT_NUMBER = 5;
const SYNCED_NATIONAL_WITHOUT_TAXES = 6;
const SYNCED_WITH_ROUNDING_ISSUES = 7;
// Generic error
const SYNC_ERROR = 10;
// Greater than 10 are all detailed errors
const INVOICE_REPEATED = 11;
const INVALID_MODEL = 12;
const GENERIC_ERROR = 13;
const TOTAL_AMOUNT_INVALID = 14;
const MAX_ATTEMPS_REACHED = 15;
const PAYMENT_SYNC_ERROR = 16;
const PAYMENT_TOO_COMPLEX = 17;
const INVALID_VAT = 18;
const PLAN_LIMIT_REACHED = 19;
const PAYMENT_DELETE_ERROR = 20;
const PAYMENT_RETRIEVAL_ERROR = 21;
const COMPANY_NIF_DISCREPANCY = 22;
const INVALID_LINES = 23;
const DELETE_ERROR = 24;
const MISSING_ORDER = 25;
const INVALID_NUMBER_DECIMALS = 26;
const ROUNDING_DISCREPANCY_ERROR = 27;
const INVALID_CURRENCY = 29;
const TAXES_PER_LINE_TOO_COMPLEX = 30;
const OPERATION_TYPE_TOO_COMPLEX = 31;
const ENTIDADES_COUNTRY_NIF_VALIDATION_ERROR_1 = 50;
const PLAN_LIMIT_REACHED_ENTITIES = 55;
const GET_COMPANIES_ERROR = 101;
const AGGREGATED_CUSTOMER_NOT_FOUND = 110;
const AGGREGATED_CUSTOMER_CANNOT_CREATE = 111;
const CUSTOMER_CANNOT_CREATE = 150;
const VIES_TIMEOUT = 120;
const INVOICE_INSERT_PERIOD_CLOSED = 130;
const INVOICE_UPDATE_NEW_PERIOD_CLOSED = 131;
const INVOICE_UPDATE_ORIGINAL_PERIOD_CLOSED = 132;
const INVOICE_PAYMENT_INSERT_PERIOD_CLOSED = 133;
const PAYMENT_METHODS_RETRIEVAL_ERROR = 210;
const GET_COUNTRIES_ERROR = 220;
const AUTH_DENIED = 401;
const HOST_UNREACHABLE = 504;
const INSUFFICIENT_RIGHTS_ERROR = 600;
const IRPF_PER_LINE_NOT_ALLOWED = 610;
const INVALID_DATE_COMPLETED = 620;
const DELETED_TAXES = 630;
const CANNOT_READ_COMPANY_DATA = 640;
// ... add more as we find them relevant for the consuming apps

// This is a special type of error that allows us to bypass the Wordpress translations and print directly
// the error message returned by the API.
// It's a compromise in order to avoid changing all the code and start from scratch the error code management system.
const API_ERROR = 999;

/**
 * Class Service
 *
 * This class represents a bridge between consuming PHP web applications and the Contasimple API
 * It makes heavy use of the underlying Swagger Codegen auto-generated code to create HTTP requests
 * and simplifies the connection process for the end developer, providing a mechanism to handle
 * the connection details only once during the class instantiation.
 */
class CSService
{
	protected $apiClient;
	protected $config;

	public function __construct($config = null)
	{
		// Basic configuration, includes only API endpoint
		if ($config != null) {
			$this->config = $config;
		} else {
			$this->config = new CSConfig();
		}

		// If the module that consumes the class has authenticated once it should have access tokens,
		// pass them down and renew them if necessary
		if ($this->config->getAccessToken() != null && $this->config->getRefreshToken() != null) {
			if ($this->config->getExpireTime() != null) {
				if (time() >= $this->config->getExpireTime()) {
					try {
						$this->renewAccessToken();
					} catch (\Exception $e) {
						throw $e;
					}
				}
			}

			// This header is needed in order for subsequent API calls to allow access once we have been authenticated
			$this->config->addDefaultHeader('Authorization', 'Bearer ' . $this->config->getAccessToken());
		}

		$this->apiClient = new ContasimpleApiClient($config);
	}

	/**
	 * Config Getter
	 *
	 * @return CSConfig
	 */
	public function getConfig()
	{
		return $this->config;
	}

	/**
	 * Config Setter
	 *
	 * @param CSConfig $config
	 */
	public function setConfig($config)
	{
		$this->config = $config;
	}

	/**
	 * Get access token
	 *
	 * Gets the OAuth 2 token, which is sent in every request that needs authentication
	 *
	 * @return string   The API generated access token if auth request was previously sent and response received, null otherwise
	 */
	public function getAccessToken()
	{
		return $this->apiClient->getConfig()->getAccessToken();
	}

	/**
	 * Set access token
	 *
	 * The access token is regenerated on each call so the consuming app will find this method useful to keep it updated in
	 * the service instantiated object once he gets the previous call response.
	 *
	 * @param $token    The new access token
	 */
	public function setAccessToken($token)
	{
		$config = $this->apiClient->getConfig();
		$config->setAccessToken($token);
		$this->apiClient = new ContasimpleApiClient($config);
	}

	/**
	 * Renew access token
	 *
	 * Since the access token expires frequently, we have to refresh it and store the new value in our config object.
	 * The refresh token is also susceptible of changing as well.
	 * This method deals with both and ensures that the API calls will be fulfilled.
	 *
	 * @throws Exception
	 */
	public function renewAccessToken()
	{
		try {
			$this->apiClient = new ContasimpleApiClient($this->config);
			$authApi = new ContasimpleAuthenticationApi($this->apiClient);
			$response = $authApi->authenticationPost("2", "refresh_token", $this->config->getClientId(), $this->config->getClientSecret(), 'offline_access', $this->config->getRefreshToken());
			$this->config->setAccessToken($response->getAccessToken());
			$this->config->setRefreshToken($response->getRefreshToken());
			$this->config->setExpireTime(time() + $response->getExpiresIn());
		} catch (ContasimpleApiException $e) {
			// It may happen that the refresh token is not valid as well, handle it for the user, try at least to reconnect again with the ApiKey
			if (empty($e->getResponseBody())) {
				if (stripos($e->getMessage(), 'Could not resolve') !== false) {
					throw new \Exception($e->getMessage(), HOST_UNREACHABLE, $e);
				} else {
					throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
				}
			} elseif ($e->getResponseBody()->errorCode == 'invalid_grant') {
				try {
					$response = $this->connect($this->config->getApiKey('apikey'));
					$this->config->setAccessToken($response->getAccessToken());
					$this->config->setRefreshToken($response->getRefreshToken());
					$this->config->setExpireTime(time() + $response->getExpiresIn());
				} catch (ContasimpleApiException $e) {
					throw new \Exception($e->getResponseBody()->errorCode, $e->getCode(), $e);
				}
			} else {
				throw new \Exception($e->getResponseBody()->errorCode, $e->getCode(), $e);
			}
		}
	}

	/**
	 * Connect to the Contasimple API
	 *
	 * Connects to CS API via authentication with provided user ApiKey
	 *
	 * @param $apikey       The ApiKey provided by CS to the end user/3rd party app
	 * @return string       JSON encoded response with access token, expiration time, refresh token, Client ID and Client Secret
	 * @throws Exception    ApiException instance with error info in responseBody field, if could not connect
	 *                      See ApiException in Swagger Codegen for more detail.
	 */
	public function connect($apikey, $userAgent = null)
	{
		try {
			$apiClient = new ContasimpleApiClient();
			$apiClient->getConfig()->setHost(URL_CS_API);
			$apiClient->getConfig()->setCurlTimeout(CURL_TIMEOUT);

			if (isset($userAgent)) {
				$apiClient->getConfig()->setUserAgent($userAgent);
			}

			$authApi = new ContasimpleAuthenticationApi($apiClient);
			$response = $authApi->authenticationPost("2", "authentication_key", null, null, 'offline_access', null, null, null, null, null, null, $apikey);

			$this->config->setApiKey('apikey', $apikey);
			$this->config->setUsername($response->getUsername());
			$this->config->setAccessToken($response->getAccessToken());
			$this->config->setRefreshToken($response->getRefreshToken());
			$this->config->setClientId($response->getClientId());
			$this->config->setClientSecret($response->getClientSecret());
			$this->config->addDefaultHeader('Authorization', 'Bearer ' . $this->config->getAccessToken());

			if (isset($userAgent)) {
				$this->config->setUserAgent($userAgent);
			}

			$this->apiClient = new ContasimpleApiClient($this->config);

			return $response;
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				if (stripos($e->getMessage(), 'Could not resolve') !== false) {
					throw new \Exception($e->getMessage(), HOST_UNREACHABLE, $e);
				} else {
					throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
				}
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, $e->getCode(), $e);
			}
		}
	}

	/**
	 * Get user available companies
	 *
	 * Gets an array of Objects of type CompanyApiModel with getters/setters available for all company attributes.
	 * Ex: getId(), getOrganizationName()
	 *
	 * @see: \Swagger\Client\Model\CompanyApiModel
	 * @return array of CompanyApiModel object types
	 * @throws Exception with error message provided by the API responseBody if could not connect, ApiException is provided as the previous exception for more detail.
	 */
	public function getCompanies()
	{
		try {
			$companiesApi = new ContasimpleMeCompaniesApi($this->apiClient);
			$companies = $companiesApi->myCompaniesGetCompanies('2')->getData();
			return $companies;
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				if (stripos($e->getMessage(), 'Could not resolve') !== false) {
					throw new \Exception($e->getMessage(), HOST_UNREACHABLE, $e);
				}
			} else {
				if (stripos($e->getResponseBody()->message, 'Authorization has been denied') !== false) {
					throw new \Exception($e->getResponseBody()->message, AUTH_DENIED, $e);
				}
			}

			throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
		}
	}

	/**
	 * Get the current company for the user
	 *
	 * The user will have at least one active company in CS, which he can change. Get the currently active one.
	 *
	 * @return \Swagger\Client\Model\CompanyApiModel
	 * @throws Exception
	 */
	public function getCurrentCompany()
	{
		try {
			$companiesApi = new ContasimpleMeCompaniesApi($this->apiClient);
			$company = $companiesApi->myCompaniesGetCurrent('2')->getData();
			return $company;
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				if (stripos($e->getMessage(), 'Could not resolve') !== false) {
					throw new \Exception($e->getMessage(), HOST_UNREACHABLE, $e);
				}
			} else {
				if (stripos($e->getResponseBody()->message, 'Authorization has been denied') !== false) {
					throw new \Exception($e->getResponseBody()->message, AUTH_DENIED, $e);
				}
			}
			// An API error not accounted for from the start, pass down the returned error by the API.
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
			}

		} catch (\Exception $e) {
			// An error for whatever other reason, the generic error can be a cul-the-sac for the CMS to display
			// in-built error messages with the internal translation mechanism of the CMS.
			throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
		}
	}

	/**
	 * Select the active company for the user
	 *
	 * Based on a previous requested list of available companies, the user can set his desired working company
	 *
	 * @param $companyId                The company ID
	 * @return AuthenticationResponse   The response from the API, will contain a new pair of access/refresh tokens as this is set at the company level
	 * @throws Exception
	 */
	public function selectCompany($companyId)
	{
		try {
			$authApi = new ContasimpleAuthenticationApi($this->apiClient); // TODO verify that this is correct, until now we didn't pass the apiClient and seemed to work fine...
			$response = $authApi->authenticationPost("2", "change_company", $this->config->getClientId(), $this->config->getClientSecret(), 'offline_access', $this->config->getRefreshToken(), null, null, null, $companyId, null, null);

			$this->config->setAccessToken($response->getAccessToken());
			$this->config->setRefreshToken($response->getRefreshToken());
			$this->config->setUsername($response->getUsername());
			$this->config->setCompanyId($companyId);
			$this->config->addDefaultHeader('Authorization', 'Bearer ' . $this->config->getAccessToken());
			$this->apiClient = new ContasimpleApiClient($this->config);

			return $response;
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
			}
		} catch (\Exception $e) {
			throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
		}
	}

	/**
	 * Update fiscal information fot the given company
	 *
	 * Calling modules can update the company info with this API method so that they match.
	 * @todo For the time being only changing the registration number/NIF is allowed.
	 *
	 * @param $companyId            The company ID
	 * @param $registrationNumber   The registration number, which in spain is the NIF/DNI
	 * @throws Exception
	 */
	public function updateCompany($companyId, $registrationNumber)
	{
		try {
			$companiesApi = new ContasimpleMeCompaniesApi($this->apiClient);
			$companiesList = $companiesApi->myCompaniesGetCompanies('2')->getData();
			foreach ($companiesList as $company) {
				if ($company->getId() == $companyId) {
					if (!empty($company->getExtraInformation())) {
						$entity = $company->getExtraInformation()->getEntity();
						$entity->setNif($registrationNumber);
						$configureModel = new ConfigureCompanyApiModel();
						$configureModel->setCountryId($company->getCountryId());
						$configureModel->setFiscalRegionId($company->getFiscalRegionId());
						$configureModel->setCompanyTypeId($company->getCompanyTypeId());
						$configureModel->setIrpfModeId($company->getIrpfModeId());
						$configureModel->setVatModeId($company->getVatTypeId());
						$configureModel->setEntity($entity);
						$companiesApi->myCompaniesConfigureCompany($configureModel, "2");
					}

					break;
				}
			}
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
			}
		} catch (\Exception $e) {
			throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
		}
	}

	/**
	 * Check if the given company needs fiscal configuration at the CS level
	 *
	 * Before a calling module can use the API to sync invoices, the CS company must have been correctly
	 * configured, for example via the provided CS wizard in the main contasimple.com site.
	 *
	 * @param $companyId    The company ID
	 * @return bool         true if needs configuration (module cannot be used), false otherwise (module can be used)
	 * @throws Exception
	 */
	public function companyRequiresConfiguration($companyId)
	{
		try {
			$companiesApi = new ContasimpleMeCompaniesApi($this->apiClient);
			$companiesList = $companiesApi->myCompaniesGetCompanies('2')->getData();//myCompaniesGetCompanies
			foreach ($companiesList as $company) {
				if ($company->getId() == $companyId) {
					return $company->getRequiresConfiguration();
				}
			}
		} catch (ContasimpleApiException $e) {
			throw new \Exception($e->getResponseBody()->message, GENERIC_ERROR, $e);
		} finally {
			return true;
		}
	}

	/**
	 * Create an invoice
	 *
	 * The calling module needs to gather the invoice information as an InvoiceApiModel object and
	 * format a valid period for the invoice, as the period is always mandatory for a CS invoice
	 *
	 * @param $period           A CS compliant period string. Ex: '2017-1T'
	 * @param $invoice          See \Swagger\Client\Model\InvoiceApiModel
	 * @return InvoiceApiModel
	 * @throws Exception        If any of the required fields is missing or validation checks are not fulfilled at the API server side
	 */
	public function createInvoice($period, $invoice)
	{
		try {
			$invoiceModel = new CreateIssuedInvoiceApiModel($invoice);
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedCreateInvoice($period, $invoiceModel, "2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			switch ($e->getResponseBody()->errorCode) {
				case 'Api_Invoices_InvalidVatPercentage':
					$errorCode = INVALID_VAT;
					break;
				case 'Invoice_RepeatedNumber':
					$errorCode = INVOICE_REPEATED;
					break;
				case 'Common_InvalidArgument':
					if (strpos($e->getResponseBody()->errorMessage, 'VAT') !== false) {
						$errorCode = INVALID_VAT;
					} elseif (strpos($e->getResponseBody()->errorMessage, 'Lines') !== false) {
						$errorCode = INVALID_LINES;
					} elseif (strpos($e->getResponseBody()->errorMessage, 'Invalid number of decimals') !== false) {
						$errorCode = INVALID_NUMBER_DECIMALS;
					} else {
						$errorCode = INVALID_MODEL;
					}
					break;
				case 'Common_PlanLimitReached':
					$errorCode = PLAN_LIMIT_REACHED;
					break;
				case 'Invoice_Insert_PeriodClosed':
					$errorCode = INVOICE_INSERT_PERIOD_CLOSED;
					break;
				case 'InvoiceLine_TaxableAmountDiscrepancy':
					$errorCode = ROUNDING_DISCREPANCY_ERROR;
					break;
				case 'entidades_country_nif_validation_error_1':
					$errorCode = ENTIDADES_COUNTRY_NIF_VALIDATION_ERROR_1;
					break;
				default:
					$errorCode = API_ERROR;
					break;
			}
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (Exception $e) {
				$errorCode = SYNC_ERROR;
				throw new \Exception($e->getMessage(), $errorCode, $e);
		}
	}

	/**
	 * Updates an existing invoice
	 *
	 * The existing invoice ID must be provided, alongside the desired new invoice parameters as a InvoiceApiModel object and the period as always
	 *
	 * @param $period       A CS compliant period string. Ex: '2017-1T'
	 * @param $invoice      See \Swagger\Client\Model\InvoiceApiModel
	 * @param $id           The existing invoice ID
	 * @return InvoiceApiModel
	 * @throws Exception    If any of the required fields is missing or validation checks are not fulfilled at the API server side
	 */
	public function updateInvoice($period, $invoice, $id)
	{
		try {
			$invoiceModel = new UpdateIssuedInvoiceApiModel($invoice);
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedUpdateInvoice($period, $id, $invoiceModel, "2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			$errorCode = SYNC_ERROR;
			switch ($e->getResponseBody()->errorCode) {
				case 'Common_InvalidArgument':
					if (preg_match('(VAT)', $e->getResponseBody()->errorMessage) === 1) {
						$errorCode = INVALID_VAT;
					} else {
						$errorCode = INVALID_MODEL;
					}
					break;

				case 'Invoice_Update_OriginalPeriodClosed':
					$errorCode = INVOICE_UPDATE_ORIGINAL_PERIOD_CLOSED;
					break;

				case 'Invoice_Update_NewPeriodClosed':
					$errorCode = INVOICE_UPDATE_NEW_PERIOD_CLOSED;
					break;
			}
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		}
	}

	/**
	 * Deletes an invoice given its ID and period
	 *
	 * @param $period   A CS compliant period string. Ex: '2017-1T'
	 * @param $id       The existing invoice ID
	 * @return int      See \Swagger\Client\Model\ApiResultInt64
	 * @throws Exception
	 */
	public function deleteInvoice($period, $id)
	{
		try {
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedDeleteInvoice($period, $id, "2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			$errorCode = DELETE_ERROR;
			switch ($e->getResponseBody()->errorCode) {
				case 'Common_InvalidArgument':
					break;
			}
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		}
	}

	/**
	 * Gets customer info based on DNI/NIF number
	 *
	 * Useful to check wheter a new client must be synced via the API
	 *
	 * @param $taxNumber            Identity Number which can be the DNI/NIF, VAT number if company has it, etc.
	 * @return null|EntityApiModel  An Entity model with customer data if it existed, or null if the customer does not exist as a CS client
	 * @throws Exception
	 */
	public function getCustomerByNIF($taxNumber)
	{
		try {
			$customersApi = new ContasimpleEntitiesCustomersApi($this->apiClient);
			$response = $customersApi->entityCustomersSearchByNif("2", $taxNumber);
			if ($response->getCount() > 0) {
				if (!empty($response->getData())) {
					return $response->getData()[0];
				} else {
					return null;
				}
			} else {
				return null;
			}
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
			}
		} catch (\Exception $e) {
			throw $e;
		}
	}

	public function getCustomerAggregated()
	{
		try {
			$customersApi = new ContasimpleEntitiesCustomersAggregatedApi($this->apiClient);
			$response = $customersApi->aggregatedCustomersEntityGet("2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				if ($e->getCode() == 404) {
					// If not configured, try to create it
					return $this->createCustomerAggregated();
				}
			}
		} catch (\Exception $e) {
			throw $e;
		}
	}

	public function createCustomerAggregated()
	{
		try {
			$customersApi = new ContasimpleEntitiesCustomersAggregatedApi($this->apiClient);
			$response = $customersApi->aggregatedCustomersEntityPost("2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			throw new \Exception($e->getResponseBody()->errorMessage, AGGREGATED_CUSTOMER_CANNOT_CREATE, $e);
		}
	}

	/**
	 * Creates a customer in CS
	 *
	 * A customer entity is always needed before being able to sync an issued invoice (to that customer)
	 *
	 * @param $customer
	 * @return \Swagger\Client\Model\EntityApiModel
	 * @throws Exception
	 */
	public function createCustomer($customer)
	{
		try {
			$customersApi = new ContasimpleEntitiesCustomersApi($this->apiClient);
			$response = $customersApi->entityCustomersPost($customer, "2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			$errorCode = CUSTOMER_CANNOT_CREATE;
			if ( ! empty( $e->getResponseBody() ) ) {
				switch ($e->getResponseBody()->errorCode) {
					case 'Common_PlanLimitReached':
						$errorCode = PLAN_LIMIT_REACHED_ENTITIES;
						break;
				}
			}
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		}
	}

	/**
	 * Send an invoice payment
	 *
	 * Invoices are not marked as pain in CS on creation time, thus this method allows to attach a payment to the invoice.
	 *
	 * @param $id
	 * @param $payment
	 * @param $period
	 * @return \Swagger\Client\Model\InvoicePaymentApiModel
	 * @throws Exception
	 */
	public function assignPayment($id, $payment, $period)
	{
		try {
			$paymentApi = new ContasimpleAccountingInvoicesIssuedPaymentsApi($this->apiClient);
			$response = $paymentApi->accountingInvoicesIssuedPaymentsCreatePayment($id, $payment, "2", $period);
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			switch ($e->getResponseBody()->errorCode) {
				case 'InvoicePayment_Insert_PeriodClosed':
					$errorCode = INVOICE_PAYMENT_INSERT_PERIOD_CLOSED;
					break;
				default:
					$errorCode = API_ERROR;
					break;
			}
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (\Exception $e) {
			throw new \Exception($e->getMessage(), PAYMENT_SYNC_ERROR, $e);
		}
	}

	/**
	 * Get invoice payments
	 *
	 * Returns a list of payments attached to the given invoice
	 *
	 * @param $idInvoice
	 * @param $period
	 * @return \Swagger\Client\Model\InvoicePaymentApiModel[]
	 * @throws Exception
	 */
	public function getInvoicePayments($idInvoice, $period)
	{
		try {
			$paymentApi = new ContasimpleAccountingInvoicesIssuedPaymentsApi($this->apiClient);
			$response = $paymentApi->accountingInvoicesIssuedPaymentsGetPayments($idInvoice, "2", $period);
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			$errorCode = PAYMENT_RETRIEVAL_ERROR;
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (\Exception $e) {
			throw new \Exception('Internal Exception', -1, $e);
		}
	}

	/*
	 * Delete payment for a given invoice
	 *
	 * CS will not mark an invoice as 'Unpaid' if the invoice details (total price) are modified via an invoice update.
	 * Thus, we need a method to delete the invoice so that we can upload payments again.
	 * This method allows to delete one payment at a time, which we can get via getInvoicePayments()
	 *
	 * @param $idInvoice    The invoice ID
	 * @param $idPayment    The ID for the desired payment to delete
	 * @param $period       The period of the invoice/payments
	 * @return \Swagger\Client\Model\ApiResultInt64
	 * @throws Exception
	 */
	public function deletePayment($idInvoice, $idPayment, $period)
	{
		try {
			$paymentApi = new ContasimpleAccountingInvoicesIssuedPaymentsApi($this->apiClient);
			$response = $paymentApi->accountingInvoicesIssuedPaymentsDeletePayment($idInvoice, $idPayment, "2", $period);
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			$errorCode = PAYMENT_DELETE_ERROR;
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (\Exception $e) {
			throw new \Exception('Internal Exception', -1, $e);
		}
	}

	public function getPaymentMethods()
	{
		try {
			$configurationPaymentMethodsApi = new ContasimpleConfigurationPaymentMethodsApi($this->apiClient);
			$response = $configurationPaymentMethodsApi->paymentMethodsConfigurationGetPaymentMethods("2");
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				if ($e->getResponseBody()->errorCode === 'Common_InsuficientRightsForService') {
					throw new \Exception($e->getResponseBody()->errorMessage, INSUFFICIENT_RIGHTS_ERROR, $e);
				} else {
					throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
				}
			}
		} catch (\Exception $e) {
			$errorCode = PAYMENT_METHODS_RETRIEVAL_ERROR;
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		}
	}

	public function getNextInvoiceNumber($period, $mask)
	{
		try {
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedGetNextInvoiceNumber($period, 2, $mask);
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				if ($e->getResponseBody()->errorCode === 'Common_InsuficientRightsForService') {
					throw new \Exception($e->getResponseBody()->errorMessage, INSUFFICIENT_RIGHTS_ERROR, $e);
				} else {
					throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
				}
			}
		} catch (\Exception $e) {
			throw new \Exception($e->getMessage(), GENERIC_ERROR, $e);
		}
	}

	public function getInvoicePDF($id, $period, $number)
	{
		try {
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedPdf($id, 2, $period);

			header("Content-Disposition: attachment; filename=\"" . basename($number . ".pdf") . "\"");
			header("Content-Type: application/octet-stream");
			header("Content-Length: " . strlen($response[0]));
			header("Connection: close");

			echo $response[0];
		} catch (ContasimpleApiException $e) {
			$errorCode = GENERIC_ERROR;
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (\Exception $e) {
			throw new \Exception('Internal Exception', -1, $e);
		}
	}

	public function getInvoicePDFasString($id, $period, $number)
	{
		try {
			$invoiceApi = new ContasimpleAccountingInvoicesIssuedApi($this->apiClient);
			$response = $invoiceApi->accountingInvoicesIssuedPdf($id, 2, $period);

			return $response[0];
		} catch (ContasimpleApiException $e) {
			$errorCode = GENERIC_ERROR;
			throw new \Exception($e->getResponseBody()->errorMessage, $errorCode, $e);
		} catch (\Exception $e) {
			throw new \Exception('Internal Exception', -1, $e);
		}
	}

	public function getCountries($all = null)
	{
		try {
			$configurationCommonApi = new ContasimpleConfigurationCommonApi($this->apiClient);
			$response = $configurationCommonApi->countriesConfigurationGetCountries("2", $all);
			return $response->getData();
		} catch (ContasimpleApiException $e) {
			if (empty($e->getResponseBody())) {
				throw new \Exception($e->getMessage(), API_ERROR, $e);
			} else {
				throw new \Exception($e->getResponseBody()->errorMessage, API_ERROR, $e);
			}
		} catch (\Exception $e) {
			$errorCode = GET_COUNTRIES_ERROR;
			throw new \Exception( $e->getResponseBody()->errorMessage, $errorCode, $e );
		}
	}
}
